
#include	"dab-constants.h"
#include	"dab_serial.h"
#ifdef	KJMP
#include	"mp2processor-kjmp.h"
#else
#include	"mp2processor-ffmpeg.h"
#endif
#include	"mp4processor.h"
#include	"deconvolve.h"
#include	"audiosink.h"
#include	"gui.h"

//
//	Interleaving is - for reasons of simplicity - done
//	inline rather than through a special class-object
static
int8_t	interleaveDelays [] = {
	     15, 7, 11, 3, 13, 5, 9, 1, 14, 6, 10, 2, 12, 4, 8, 0};
//
//
//	fragmentsize == Length * CUSize
	dabSerial::dabSerial	(uint8_t dabModus,
	                         int16_t fragmentSize,
	                         int16_t bitRate,
	                         int16_t uepFlag,
	                         int16_t protLevel,
	                         int16_t tracebackLevel,
	                         RadioInterface *mr,
	                         audioSink *as) {
int32_t i, j;
	this	-> dabModus		= dabModus;
	this	-> fragmentSize		= fragmentSize;
	this	-> bitRate		= bitRate;
	this	-> uepFlag		= uepFlag;
	this	-> protLevel		= protLevel;
	this	-> tracebackLevel	= tracebackLevel;
	this	-> myRadioInterface	= mr;
	this	-> myAudioSink		= as;
	outV			= new uint8_t [bitRate * 24];
	interleaveData		= new int16_t *[fragmentSize]; // the size
	for (i = 0; i < fragmentSize; i ++) {
	   interleaveData [i] = new int16_t [16];
	   for (j = 0; j < 16; j ++)
	      interleaveData [i][j] = 0;
	}
	countforInterleaver	= 0;

	uepProcessor		= NULL;
	eepProcessor		= NULL;
	if (uepFlag == 0)
	   uepProcessor	= new uep_deconvolve (bitRate,
	                                      protLevel,
	                                      tracebackLevel);
	else
	   eepProcessor	= new eep_deconvolve (bitRate,
	                                      protLevel,
	                                      tracebackLevel);
//
	
	if (dabModus == DAB) 
	   our_dabProcessor = new mp2Processor (myRadioInterface,
	                                        myAudioSink, bitRate);
	else
	if (dabModus == DAB_PLUS) 
	   our_dabProcessor = new mp4Processor (myRadioInterface,
	                                        myAudioSink, bitRate);
	else		// cannot happen
	   our_dabProcessor = new dabProcessor ();

	myAudioSink	-> restart	();
}

	dabSerial::~dabSerial	(void) {
int16_t	i;
//	myAudioSink	-> stop ();
	if (uepProcessor != NULL)
	   delete uepProcessor;
	if (eepProcessor != NULL)
	   delete eepProcessor;
	delete our_dabProcessor;
	delete[]	outV;
	for (i = 0; i < fragmentSize; i ++) 
	   delete[]  interleaveData [i];
	delete[] interleaveData;
}

int32_t	dabSerial::process	(int16_t *Data, int16_t cnt) {
int16_t	i, j;
uint8_t	shiftRegister [9];
//
//	first: interleaving
	for (i = 0; i < fragmentSize; i ++) {
	   interleaveData [i][interleaveDelays [i & 017]] = Data [i];
	   Data [i] = interleaveData [i] [0];
//	and shift
	   memmove (&interleaveData [i][0],
	            &interleaveData [i][1],
	            interleaveDelays [i & 017] * sizeof (int16_t));
	}
//
//	only continue when de-interleaver is filled
	if (countforInterleaver < 15) {
	   ++countforInterleaver;
	   return 32768;
	}
//
	if (uepFlag == 0)
	   uepProcessor -> deconvolve (Data, fragmentSize, outV);
	else
	   eepProcessor -> deconvolve (Data, fragmentSize, outV);
//
//	and the inline energy dispersal
//	and for now, write out
	memset (shiftRegister, 1, 9);
	for (i = 0; i < bitRate * 24; i ++) {
	   uint8_t b = shiftRegister [8] ^ shiftRegister [4];
	   for (j = 8; j > 0; j--)
	      shiftRegister [j] = shiftRegister [j - 1];
	   shiftRegister [0] = b;
	   outV [i] ^= b;
	}

	our_dabProcessor	-> addtoFrame (outV, 24 * bitRate);
	return 32768;
}
//
void	dabSerial::stopRunning (void) {
//	fprintf (stderr, "stopping\n");
//	myAudioSink	-> stop ();
}

