#
#include	<cstring>
#include	"msc_handler.h"
#include	<stdlib.h>
#include	"jff-include.h"
#include	"gui.h"
#include	"dab_virtual.h"
#include	"dab_serial.h"
#include	"dab_concurrent.h"
#include	<unistd.h>
//
//	Driver program for processing the MSC.
//	Three operations here (apart from selecting
//	the local frame in the MSC vector)
//	1. deinterleaving
//	2. deconvolution (including depuncturing)
//	3. energy dispersal
//	4. in case of DAB: creating MP2 packets
//	5. in case of DAB+: assembling superframes and creating MP4 packets
//
//	The selected service(s) is (are) to be found in the
//	ficParameters.

#define	CUSize	(4 * 16)
//	Note CIF counts from 0 .. 3
//
		mscHandler::mscHandler	(RadioInterface *mr,
	                                 int16_t tracebackLevel,
	                                 audioSink	*sink,
	                                 uint8_t concurrent) {
		myRadioInterface	= mr;
	        if ((tracebackLevel < 2) || (tracebackLevel > 30))
	           tracebackLevel = 2;
	        this -> tracebackLevel = tracebackLevel;
	        concurrencyOn 		= concurrent != 0;
	        cifVector		= new int16_t [55296];
	        cifCount		= 0;	// msc blocks in CIF
	        blkCount		= 0;
	        dabHandler		= new dabVirtual;
	        newChannel		= false;
	        currentChannel		= -1;
	        dabModus		= 0;
	        our_audioSink		= sink;
}

		mscHandler::~mscHandler	(void) {
	delete[]  cifVector;
	delete	dabHandler;
}

void	mscHandler::setChannel (int16_t subchId,
	                        int16_t uepFlag,
	                        int16_t	startAddr,
	                        int16_t	Length,
	                        int16_t	protLevel,
	                        int16_t	bitRate,
	                        int16_t	ASCTy,
	                        int16_t	language,
	                        int16_t	type) {
	newChannel	= true;
	currentChannel	= subchId;
	new_uepFlag	= uepFlag;
	new_startAddr	= startAddr;
	new_Length	= Length;
	new_protLevel	= protLevel;
	new_bitRate	= bitRate;
	new_language	= language;
	new_type	= type;
	new_dabModus	= ASCTy == 077 ? DAB_PLUS : DAB;
	fprintf (stderr, "Preparations for channel select\n");
}
//
//
//	add blocks. First is (should be) block 5, last is (should be) 76
void	mscHandler::process_mscBlock	(int16_t *fbits,
	                                 int16_t blkno) {
int16_t	currentblk	= (blkno - 5) % 18;
int16_t	*myBegin;

	if (currentChannel == -1)
	   return;
//
//	we only change channel at the start of a new frame!!!
	if ((currentblk == 0) && newChannel) {
	   newChannel	= false;
	   dabHandler -> stopRunning ();
	   delete dabHandler;

	   if (concurrencyOn)
	      dabHandler = new dabConcurrent (new_dabModus,
	                                      new_Length * CUSize,
	                                      new_bitRate,
	                                      new_uepFlag,
	                                      new_protLevel,
	                                      tracebackLevel,
	                                      myRadioInterface,
	                                      our_audioSink);
	   else
	      dabHandler = new dabSerial (new_dabModus,
	                                  new_Length * CUSize,
	                                  new_bitRate,
	                                  new_uepFlag,
	                                  new_protLevel,
	                                  tracebackLevel,
	                                  myRadioInterface,
	                                  our_audioSink);
	                                  
	   startAddr	= new_startAddr;
	   Length	= new_Length;
	   protLevel	= new_protLevel;
	   bitRate	= new_bitRate;
	}
//
//	and the normal operation is:
//	We might optimize this by realizing that the exact segment
//	we take from the CIF is known.
	memcpy (&cifVector [currentblk * 3072],
	                    fbits, 3072 * sizeof (int16_t));
	if (++blkCount < 18) 
	   return;
//
//	OK, now we have a full CIF
	blkCount	= 0;
	cifCount	= (cifCount + 1) & 03;
	myBegin		= &cifVector [startAddr * CUSize];
//
//	Here we move the vector to be processed to a
//	separate task or separate function, depending on
//	the settings in the ini file, we might take advantage of multi cores
	(void) dabHandler -> process (myBegin, Length * CUSize);
}
//
//
uint8_t	mscHandler::getMode	(void) {
	return new_dabModus;
}

int16_t	mscHandler::getChannel	(void) {
	return currentChannel;
}

int16_t	mscHandler::getLanguage (void) {
	return new_language;
}

int16_t	mscHandler::getType (void) {
	return new_type;
}

void	mscHandler::stopProcessing (void) {
	currentChannel = -1;
}

