/*
 *
 */
#include "viterbi.h"
/*
 *	The viterbi algorithm has been recoded for fitness in the
 *	JSDR framwork.
 *	a specialised version for the DAB reception
 */
#define		K	7
#define		Poly1	0133
#define		Poly2	0171
#define		Poly3	0145
#define		Poly4	0133
#define		MEM_SIZE	0400

	viterbi::viterbi (int16_t traceLength) {
int16_t	i, j;

	v_chunksize		= 1;
	v_traceLength		= traceLength;
	NumofStates		= 1 << (K - 1);
	count			= 0;
//
	transCosts		= new int32_t *[MEM_SIZE];
	history			= new int16_t *[MEM_SIZE];
	sequence		= new int16_t  [MEM_SIZE];
//
	for (i = 0; i < MEM_SIZE; i++) {
	   transCosts [i]	= new int32_t [NumofStates];
	   history [i]		= new int16_t [NumofStates];
	   sequence [i] 	= 0;
	   for (j = 0; j < NumofStates; j ++) {
	      transCosts [i][j] = 0;
	      history [i][j] = 0;
	   }
	}
//
//	These tables give a mapping from (state * bit * Poly -> outputbit)
	poly1_table	= new uint8_t [2 * NumofStates];
	for (i = 0; i < 2; i ++)
	   for (j = 0; j < NumofStates; j ++)
	      poly1_table [i * NumofStates + j] = bitFor (j, Poly1, i);

	poly2_table	= new uint8_t [2 * NumofStates];
	for (i = 0; i < 2; i ++)
	   for (j = 0; j < NumofStates; j ++)
	      poly2_table [i * NumofStates + j] = bitFor (j, Poly2, i);

	poly3_table	= new uint8_t [2 * NumofStates];
	for (i = 0; i < 2; i ++)
	   for (j = 0; j < NumofStates; j ++)
	      poly3_table [i * NumofStates + j] = bitFor (j, Poly3, i);

	poly4_table	= new uint8_t [2 * NumofStates];
	for (i = 0; i < 2; i ++)
	   for (j = 0; j < NumofStates; j ++)
	      poly4_table [i * NumofStates + j] = bitFor (j, Poly4, i);

	table5		= new uint8_t [2 * NumofStates];
	for (i = 0; i < 2 * NumofStates; i ++)
	   table5 [i] = (poly1_table [i] << 3) |
	                (poly2_table [i] << 2) |
	                (poly3_table [i] << 1) |
	                poly4_table [i];
	globalPtr = 0;
	bitBufferIndex	= 0;
}

	viterbi::~viterbi () {
int32_t i;
	for (i = 0; i < MEM_SIZE; i++) {
	   delete[] transCosts [i];
	   delete[] history [i];
	}

	delete[]	sequence;
	delete[]	transCosts;
	delete[]	history;
	delete		poly1_table;
	delete		poly2_table;
	delete		poly3_table;
	delete		poly4_table;
}
//
//	shift x0, x1, x2, x3
int16_t	viterbi::viterbi_shiftbit (int16_t bit, int32_t *metric) {

	bitBuffer [bitBufferIndex] = bit;
	if (++ bitBufferIndex > 3) {
	   bitBufferIndex = 0;
	   return viterbi_decode (bitBuffer, metric);
	}

	return -1;
}
//
//	sym [0] is the first read bit, i.e. x0
int16_t	viterbi::viterbi_decode (int16_t *sym, int32_t *metric) {
uint16_t	currptr, prevptr;
uint16_t	cState, bestState;
uint16_t 	i, j;
uint16_t	prev_0, prev_1;
int32_t 	costs_0, costs_1;
int32_t		minimalCosts;

	currptr = globalPtr;
	prevptr = (globalPtr - 1) & (MEM_SIZE - 1);

	for (cState = 0; cState < NumofStates; cState ++) {
	   uint8_t entrybit = (cState >= NumofStates / 2) ? 1 : 0;
/*
 *	we know that each state has two predecessors, to be obtained
 *	by shifting the "youngest" (i.e the leftmost) bit out
 *	and adding "oldest" bits zero or one in
 */
	   prev_0 = ((cState << 1) + 00) & (NumofStates - 1);
	   prev_1 = ((cState << 1) + 01) & (NumofStates - 1);
/*
 *	we compute the minimal costs, based on the costs of the
 *	prev states, and the additional costs of arriving from
 *	the previous state to the current state with the symbol "sym"
 */
	   costs_0 = transCosts [prevptr][prev_0] +
	             costsFor (prev_0, entrybit, sym);
	   costs_1 = transCosts [prevptr][prev_1] +
	             costsFor (prev_1, entrybit, sym);
	   if (costs_0 < costs_1) {
	      transCosts [currptr][cState] = costs_0;
	      history [currptr][cState] = prev_0;
	   } else {
	      transCosts [currptr][cState] = costs_1;
	      history [currptr][cState] = prev_1;
	   }
	}
/*
 *	OK, cost matrix is updated
 */
	globalPtr = (globalPtr + 1) & (MEM_SIZE - 1);
//	The first v_traceLength steps, it does not make much sense
//	to do a traceLength. 
	if ((++ count > v_traceLength) && 
	                   ((globalPtr % v_chunksize) == 0)) {// ok, let us decode
	   minimalCosts	= INT_MAX;
	   bestState	= 0;
	   for (i = 0; i < NumofStates; i++) {
	      if (transCosts [currptr][i] < minimalCosts) {
	         minimalCosts = transCosts [currptr][i];
	         bestState = i;
	      }
	   }
	   return traceBack (currptr, bestState, metric);
	}

	if (transCosts [currptr][0] > INT_MAX / 2) {
	   for (i = 0; i < MEM_SIZE; i++)
	      for (j = 0; j < NumofStates; j++)
	         transCosts [i][j] -= INT_MAX / 2;
	}

	if (transCosts [currptr][0] < INT_MIN / 2) {
	   for (i = 0; i < MEM_SIZE; i++)
	      for (j = 0; j < NumofStates; j++)
	         transCosts [i][j] += INT_MAX / 2;
	}

	return -1;
}
/*
 *	Trace back 'traceBack' steps, starting from the best state
 *	and decode the "chunksize" amount of bits
 */
int16_t viterbi::traceBack (uint16_t p, int16_t beststate, int32_t *metric) {
uint16_t prev;		// pointer
int32_t	localCosts;
int16_t	i;
uint16_t	c;		// the result

	sequence [p] = beststate;
	for (i = 0; i < v_traceLength; i++) {
	   prev = (p - 1) & (MEM_SIZE - 1);
	   sequence [prev] = history [p][sequence[p]];
	   p = prev;
	}

	localCosts = transCosts [p][sequence[p]];
//
// Decode 'chunksize' bits
//
//	Note that we need the entrybit of the successor of the
//	"oldest" state in the trace
	c = 0;
	for (i = 0; i < v_chunksize; i++) {
	   uint8_t newBit = (sequence [p] >= NumofStates / 2) ? 01 : 00;
	   c = (c << 1) | newBit;
	   p = (p + 1) & (MEM_SIZE - 1);
	}

	localCosts = transCosts [p][sequence [p]] - localCosts;
	if (metric)
	   *metric = localCosts;
	return c;
}

int16_t	viterbi::viterbi_set_chunksize (int16_t chunksize) {
	if ((chunksize < 1) || (chunksize > v_traceLength))
	   return -1;

	v_chunksize = chunksize;
	return 0;
}

void	viterbi::viterbi_reset (int16_t traceLength, int16_t chunksize) {
int16_t i;

	for (i = 0; i < MEM_SIZE; i++) {
	   memset (transCosts [i], 0, NumofStates * sizeof (int32_t));
	   memset (history [i], 0, NumofStates * sizeof (int16_t));
	}

	globalPtr = 0;
	count	= 0;
	v_traceLength	= traceLength;
	v_chunksize	= chunksize;
}

//
//
//	Note that the soft bits are such that
//	they are int16-t -255 -> (bit)1, +255 -> (bit)0
//	sym [0] is the "first" bit
int16_t	viterbi::costsFor (uint16_t State, uint8_t bit, int16_t *sym) {
int32_t		res	= 0;
uint8_t		targetBit	= 0;
//
//	The "switch" turned out to be far less efficient than
//	the "if ... else"
//	switch (table5 [bit * NumofStates + State] & 017) {
//	   default:	// cannot happen
//	   case 0:	// 0 0 0 0
//	      return 1204 - sym [0] - sym [1] - sym [2] - sym [3];
//
//	   case 1:	// 0 0 0 1
//	      return 1204 - sym [0] - sym [1] - sym [2] + sym [3];
//
//	   case 2:	// 0 0 1 0
//	      return 1204 - sym [0] - sym [1] + sym [2] - sym [3];
//
//	   case 3:	// 0 0 1 1
//	      return 1204 - sym [0] - sym [1] + sym [2] + sym [3];
//
//	   case 4:	// 0 1 0 0
//	      return 1204 - sym [0] + sym [1] - sym [2] - sym [3];
//
//	   case 5:	// 0 1 0 1
//	      return 1204 - sym [0] + sym [1] - sym [2] + sym [3];
//
//	   case 6:	// 0 1 1 0
//	      return 1204 - sym [0] + sym [1] + sym [2] - sym [3];
//
//	   case 7:	// 0 1 1 1
//	      return 1204 - sym [0] + sym [1] + sym [2] + sym [3];
//
//	   case 010:	// 1 0 0 0
//	      return 1204 + sym [0] - sym [1] - sym [2] - sym [3];
//
//	   case 011:	// 1 0 0 1
//	      return 1204 + sym [0] - sym [1] - sym [2] + sym [3];
//
//	   case 012:	// 1 0 1 0
//	      return 1204 + sym [0] - sym [1] + sym [2] - sym [3];
//
//	   case 013:	// 1 0 1 1
//	      return 1204 + sym [0] - sym [1] + sym [2] + sym [3];
//
//	   case 014:	// 1 1 0 0
//	      return 1204 + sym [0] + sym [1] - sym [2] - sym [3];
//
//	   case 015:	// 1 1 0 1
//	      return 1204 + sym [0] + sym [1] - sym [2] + sym [3];
//
//	   case 016:	// 1 1 1 0
//	      return 1204 + sym [0] + sym [1] + sym [2] - sym [3];
//
//	   case 017:	// 1 1 1 1
//	      return 1204 + sym [0] + sym [1] + sym [2] + sym [3];
//	}

//
//	sym [0] -> 256 is a 0
//	sym [0] -> -256 is a 1
	targetBit = poly1_table [bit * NumofStates + State];
	if (targetBit == 0)
	   res += 256 - sym [0];
	else
	   res += 256 + sym [0];

	targetBit = poly2_table [bit * NumofStates + State];
	if (targetBit == 0)
	   res += 256 - sym [1];
	else
	   res += 256 + sym [1];

	targetBit = poly3_table [bit * NumofStates + State];
	if (targetBit == 0)
	   res += 256 - sym [2];
	else
	   res += 256 + sym [2];

	targetBit = poly4_table [bit * NumofStates + State];
	if (targetBit == 0)
	   res += 256 - sym [3];
	else
	   res += 256 + sym [3];

	return res;
}
/*
 *	as an aid, we give a function "bitFor" that, given
 *	the register state, the polynome and the bit to be inserted
 *	returns the bit coming from the engine
 */
uint8_t	viterbi::bitFor (uint16_t state, uint16_t poly, uint8_t bit) {
uint16_t Register;
uint8_t	resBit = 0;
int16_t	i;
//
//	the register after shifting "bit" in would be:
	Register = bit == 0 ? state : (state + NumofStates);
	Register &= poly;
/*
 *	now for the individual bits
 */
	for (i = 0; i <= K; i++) {
	   resBit ^= (Register & 01);
	   Register >>= 1;
	}

	return resBit;
}

