#
/*
 *    Copyright (C) 2010,2011, 2012, 2013
 *    Jan van Katwijk (J.vanKatwijk@gmail.com)
 *    JFF Consultancy
 *
 *    This file is part of the  SDR-J  series.
 *    Many of the ideas as implemented in the SDR-J are derived from
 *    other work, made available through the GNU general Public License. 
 *    All copyrights of the original authors are recognized.
 *
 *    SDR-J is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    SDR-J is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with SDR-J; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include	<Qt>
#include	<QSettings>
#include	<QMessageBox>
#include	<QFileDialog>
#include	<QDebug>
#include	<qwt_color_map.h>
#include	<qwt_plot_spectrogram.h>
#include	<qwt_scale_widget.h>
#include	<qwt_scale_draw.h>
#include	<qwt_plot_zoomer.h>
#include	<qwt_plot_panner.h>
#include	<qwt_plot_layout.h>
#include	<QDateTime>
#include	<QFile>
#include	"filereader.h"
#ifndef __MINGW32__
#include	"dabstick.h"
#else
#include	"dabstick-dll.h"
#endif
#include	"gui.h"
#include	"fm_demodulator.h"
#include	"fm_processor.h"
#include	"rds_decoder.h"
#include	"decimator.h"
#include	"fft-scope.h"
#include	"scope.h"
#include	"oscillator.h"
#include	"fm-constants.h"
#include	<QInputDialog>
#include	"pa_writer.h"
#ifdef __MINGW32__
#include	<iostream>
#include	<windows.h>
#endif
//
//	Radio devices supported directly:
#define	REPLAY		0102
#define	DABSTICK	0103
#define	DABSTICK_DLL	0104

//
//	Processing modes
#define	IDLE		0100
#define	PAUSED		0101
#define	RUNNING		0102

static
//int16_t	delayTable [] = {15, 13, 11, 10, 9, 8, 7, 5, 3, 2, 1};
int16_t	delayTable [] = {1, 3, 5, 7, 9, 10, 15};
#define delayTableSize	((int)(sizeof (delayTable) / sizeof (int16_t)))
/*
 *	We use the creation function merely to set up the
 *	user interface and make the connections between the
 *	gui elements and the handling agents. All real action
 *	is embedded in actions, initiated by gui buttons
 */
	RadioInterface::RadioInterface (QSettings	*Si,
	                                int32_t		audioRate,
	                                KeyboardFilter	*myFilter,
	                                QWidget		*parent): QDialog (parent) {
int16_t	i;
	filter				= myFilter;
	connect	(filter, SIGNAL (KeyPressed (int)),
	         this, SLOT (getKeyStroke (int)));

	setupUi (this);
	ISettings			= Si;
	this		-> fmRate	= 240000;
	this		-> dabRate	= 10 * this -> fmRate;
	this		-> audioRate	= 48000;

	this		-> displaySize	= 512;
	this		-> spectrumSize	= 1024;
	this		-> rasterSize	= 50;
	this		-> repeatRate	= 10;
	this		-> averageCount	= 5;
	if (ISettings != NULL) {
	   this		-> displaySize	=
                             ISettings -> value ("displaySize",
	                                          displaySize). toInt ();
	   if ((displaySize & (displaySize - 1)) != 0)
	      displaySize = 256;

	   this		-> spectrumSize =
	                     ISettings -> value ("spectrumSize",
	                                          spectrumSize). toInt ();
	   this		-> rasterSize	=
	                     ISettings -> value ("rasterSize",
	                                          rasterSize). toInt ();
	   this		-> repeatRate	= 
	                     ISettings -> value ("repeatRate",
	                                          repeatRate). toInt ();
	   this		-> averageCount	 =
	                     ISettings -> value ("averageCount",
	                                          averageCount). toInt ();
	   this		-> audioRate	=
	                     ISettings	-> value ("audioRate",
	                                          48000). toInt ();
	   this		-> fmRate	=
	                     this -> audioRate *
	                     ISettings -> value ("fmRate_Multiplier",
	                                                     4). toInt ();
	   this		-> dabRate	=
	                     this -> fmRate *
	                     ISettings -> value ("dabRate_Multiplier",
	                                                     5). toInt ();
	}

//
//	just a sanity check
	if ((this -> audioRate != 24000 ) &&
	      (this -> audioRate != 44100) && (this -> audioRate != 48000)) {
	   this -> audioRate	= 48000;
	   this -> fmRate	= 192000;
	   this -> dabRate	= 5 * this -> fmRate;
	}
	if ((this -> fmRate < 150000) ||
	    (this -> dabRate < 900000) ||
	    (this -> dabRate > 3200000)) {
	   this -> audioRate	= 48000;
	   this -> fmRate	= 192000;
	   this -> dabRate	= 5 * this -> fmRate;
	}

	this		-> workingRate	= this ->  audioRate;

	displayBuffer			= new double [this -> displaySize];
	memset (displayBuffer, 0, displaySize * sizeof (double));

	this		-> omega	= 1.0f/(this -> fmRate) * (2 * M_PI);

	this		-> mySinCos	= new SinCos (this -> fmRate);
//
//	we need (very) high latency
	our_audioSink		= new paWriter (this -> audioRate, 
	                                        VERYHIGHLATENCY);
	outTable		= new int16_t
	                             [our_audioSink -> numberofDevices () + 1];
	for (i = 0; i < our_audioSink -> numberofDevices (); i ++)
	   outTable [i] = -1;

	if (!setupSoundOut (streamOutSelector,
	                    our_audioSink,
	                    this -> audioRate,
	                    outTable)) {
	   fprintf (stderr, "Cannot open any output device\n");
	   abortSystem (33);
	}
	runMode			= IDLE;
//
/*
 *	we also setup the basis for the two displays, overlapped
 *	with the spectrograms, and the separate IQ display
 */
	lfBuffer	= new RingBuffer<DSPCOMPLEX> (16384);
	LFScope	= new fft_scope (lfscope,
	                         this	-> displaySize,
	                         1,		// Hz
	                         this	-> rasterSize,
	                         4 * this -> displaySize,
	                         this	-> fmRate,
	                         this	-> repeatRate);
	LFScope	-> SelectView (SPECTRUM_MODE);
	LFAverager	= false;
	LFScope		-> setAverager	(LFAverager);

	connect (LFScope,
	         SIGNAL (clickedwithLeft (int)),
	         this,
	         SLOT (Zoom_in (int)));
	connect (LFScope,
	         SIGNAL (clickedwithRight (int)),
	         this,
	         SLOT (Zoom_out (int)));

	hfBuffer	= new RingBuffer<double> (2 * displaySize);
	HFScope		= new Scope (hfscope,
	                             this -> displaySize,
	                             this -> rasterSize);
	HFviewMode	= SPECTRUM_MODE;
	HFScope	-> SelectView (SPECTRUM_MODE);
	connect (HFScope,
	         SIGNAL (clickedwithLeft (int)),
	         this,
	         SLOT (AdjustFrequency (int)));
	setAmplification (spectrumAmplitudeSlider -> value ());
	ClearPanel		();
/*
 *	he does the connections from the gui buttons, sliders etc
 */
	localConnects		();
//
//	The filters with initial - useless - values
	fmFilter		= new LowPassFIR (21,
	                                          0.95 * this -> fmRate / 2,
	                                          this -> fmRate);

	dumping			= FALSE;
	MyRig			= new virtualInput ();
	deviceSelect		-> setCurrentIndex (0);
//
//	Create a timer for dealing with a non acting (or reacting) user
//	on the keyboard
	lcd_timer		= new QTimer ();
	lcd_timer		-> setSingleShot (TRUE);	// single shot
	lcd_timer		-> setInterval (5000);
	connect (lcd_timer,
	         SIGNAL (timeout ()),
	         this,
	         SLOT (lcd_timeout ()));
	key_timer		= new QTimer ();
	key_timer		-> setSingleShot (TRUE);
	key_timer		-> setInterval (5000);
	connect (key_timer,
	         SIGNAL (timeout ()),
	         this,
	         SLOT (key_timeout ()));
//
//	a timer for autoincrement/decrement of the tuning
	autoIncrement_timer	= new QTimer ();
	autoIncrement_timer	-> setSingleShot (TRUE);
	autoIncrement_timer	-> setInterval (5000);
	connect (autoIncrement_timer,
	         SIGNAL (timeout ()),
	         this,
	         SLOT (autoIncrement_timeout ()));

	log_timer		= new QTimer ();
	log_timer		-> setSingleShot (true);
	log_timer		-> setInterval (5000);
	connect (log_timer,
	         SIGNAL (timeout ()),
	         this,
	         SLOT (log_timeout ()));
/*
 *	just as a gadget, we show the actual time
 */
	display_timer		= new QTimer ();
	display_timer		-> setInterval (1000);
	connect (display_timer,
	         SIGNAL (timeout ()),
	         this,
	         SLOT (updateTimeDisplay ()));
//
//	Note that setting parameters for the decoders has
//	to wait until the decoders are instantiated!!!
	myFMprocessor		= new fmProcessor (MyRig,
	                                           this,
	                                           our_audioSink,
	                                           this -> dabRate,
	                                           this -> fmRate,
	                                           this -> workingRate,
	                                           this -> audioRate,
	                                           displaySize,
	                                           spectrumSize,
	                                           averageCount,
	                                           repeatRate,
	                                           hfBuffer,
	                                           lfBuffer,
	                                           mySinCos);
	lcd_fmRate		-> display ((int)this -> fmRate);
	lcd_dabRate		-> display ((int)this -> dabRate);
	lcd_OutputRate		-> display ((int)this -> audioRate);
	char	*s = (char *)alloca (256);
	sprintf (s, "sdrJ-FM-V%s", CURRENT_VERSION);
	systemindicator		-> setText (QString (s));
	display_timer		-> start (1000);
//
	KeyhandlingMode		= NORMAL;
	currentPIcode		= 0;
	frequencyforPICode	= 0;
//	all elements seem to exist: set the control state to
//	its default values
	setInitialControlState	();
//	if we have a ".ini" file, we overrule the initial control state
	if ((ISettings != NULL))
	   setControlState	(ISettings);
}

	RadioInterface::~RadioInterface () {
	delete		HFScope;
	delete		LFScope;

	delete		myFMprocessor;
	delete		lcd_timer;
	delete		key_timer;
	delete		autoIncrement_timer;
	delete		log_timer;
	delete		display_timer;
	delete		MyRig;
}
//
//	Eventually, we want that each individual element
//	of the control state can be overruled. We start
//	with showing this state through default settings
//	This can be overruled when an inifile is
//	provided.
void	RadioInterface::setInitialControlState	(void) {

	logFile			= NULL;

	pauseButton		-> setText (QString ("Pause"));

	dumpButton		-> setText ("dumpOn");
	dumping			= false;

	incrementingFlag ->
	                 setStyleSheet ("QLabel {background-color:blue}");
	incrementingFlag -> setText (" ");
	IncrementIndex		= 0;
	outputDevice		= our_audioSink	-> invalidDevice ();

	min_loop_frequency	= MIN_LOOP_FREQUENCY;
	max_loop_frequency	= MAX_LOOP_FREQUENCY;
	set_fm_increment	(fm_increment		-> value ());
	set_minimum		(minimumSelect		-> value ());
	set_maximum		(maximumSelect		-> value ());
	setAmplification	(spectrumAmplitudeSlider -> value ());
	setAttenuation		(attenuationSlider	 -> value ());
	setIQBalance		(IQbalanceSlider	-> value ());
	
	setInputMode		(inputModeSelect	-> currentText ());
	setVolume		(VolumeSlider		-> value ());
//
//	now setting the parameters for the fm decoder
	setfmBandwidth		(fmFilterSelect		-> currentText ());
	setfmBandwidth		(fmFilterDegree		-> value ());
	setfmMode		(fmMode			-> currentText ());
	setfmRdsSelector	(fmRdsSelector		-> currentText ());
	setfmDecoder		(fmDecoder		-> currentText ());
	setfmView		(fmView			-> currentText ());
	setfmChannelSelector	(fmChannelSelect	-> currentText ());
	setfmDeemphasis		(fmDeemphasisSelector	-> currentText ());
	setfmStereoSlider	(fmStereoSlider		-> value ());
	setfmLFcutoff		(fmLFcutoff		-> currentText ());
	setLogging		(logging		-> currentText ());

	setDevice		(deviceSelect		-> currentText ());

	dongleOffset		= 0;
	freqCorrectingValue	= 0;	// in ppm
}
//
//	We have the possibility of creating a ".ini" file
void	RadioInterface::dumpControlState	(QSettings *s) {

	if (s == NULL)
	   return;

	s	-> setValue ("displaySize", displaySize);
	s	-> setValue ("spectrumSize", spectrumSize);
	s	-> setValue ("rasterSize", rasterSize);
	s	-> setValue ("averageCount", averageCount);
	s	-> setValue ("audioRate", audioRate);
	s	-> setValue ("fmRate_Multiplier", fmRate / audioRate);
	s	-> setValue ("dabRate_Multiplier", dabRate / fmRate);

	s	-> setValue ("repeatRate", repeatRate);

	s	-> setValue ("fm_increment",
	                               fm_increment -> value ());
	s	-> setValue ("spectrumAmplitudeSlider",
	                               spectrumAmplitudeSlider -> value ());
	s	-> setValue ("attenuationSlider",
	                               attenuationSlider -> value ());
	s	-> setValue ("IQbalanceSlider",
	                               IQbalanceSlider	-> value ());
	s	-> setValue ("inputModeSelect",
	                               inputModeSelect	-> currentText ());
	s	-> setValue ("VolumeSlider",
	                               VolumeSlider	-> value ());
//
//	now setting the parameters for the fm decoder
	s	-> setValue ("fmFilterSelect", 
	                               fmFilterSelect	-> currentText ());
	s	-> setValue ("fmFilterDegree",
	                               fmFilterDegree	-> value ());

	s	-> setValue ("fmMode",
	                               fmMode 		-> currentText ());
	s	-> setValue ("fmDecoder",
	                               fmDecoder	-> currentText ());
	s	-> setValue ("fmRdsSelector",
	                               fmRdsSelector	-> currentText ());
	s	-> setValue ("fmView",
	                               fmView		-> currentText ());
	s	-> setValue ("fmChannelSelect",
	                               fmChannelSelect	-> currentText ());
	s	-> setValue ("fmDeemphasisSelector",
	                               fmDeemphasisSelector -> currentText ());
	s	-> setValue ("fmStereoSlider",
	                               fmStereoSlider	-> value ());
	s	-> setValue ("fmLFcutoff",
	                               fmLFcutoff	-> currentText ());
	s	-> setValue ("logging",
	                               logging		-> currentText ());
	s	-> setValue ("deviceSelect", 
	                               deviceSelect -> currentText ());
	s	-> setValue ("streamOutSelector",
	                               streamOutSelector -> currentText ());
//
//	in the ini file, we specify frequency in KHz
	s	-> setValue ("frequency",
	                               (int)CurrentFreq / Khz (1));
	s	-> setValue ("streamOutSelector",
	                               streamOutSelector -> currentText ());
	s	-> setValue ("IncrementIndex",
	                               IncrementIndex);

	s	-> setValue ("min_loop_frequency",
	                               min_loop_frequency);
	s	-> setValue ("max_loop_frequency",
	                               max_loop_frequency);
	s	-> setValue ("dongleOffset_Khz", dongleOffset / KHz (1));
	s	-> setValue ("dongleOffset_Hz", dongleOffset % KHz (1));
	s	-> setValue ("freqCorrector", 
	                              freqCorrector -> value ());
}
//
//	This does the trick: if an ".ini" file is specified,
//	the specified settings are restored. Note that
//	the ini file does not need to provide all settings,
void	RadioInterface::setControlState	(QSettings *s) {
QString	h;
int	k;
double	f;

	if (s == NULL)
	   return;

	k	= s -> value ("dongleOffset_Khz", 0). toInt ();
	k	=  KHz (k) + s -> value ("dongleOffset_Hz", 0). toInt ();
	dongleOffset	= k;

	k	= s -> value ("freqCorrector", 0). toInt ();
	freqCorrector		-> setValue (k);	// will signal

	k	= s -> value ("fm_increment", 100). toInt ();
	fm_increment		-> setValue (k);	// will signal

	IncrementIndex	= s -> value ("IncrementIndex", 0). toInt ();
	min_loop_frequency	=
	 s -> value ("min_loop_frequency", MIN_LOOP_FREQUENCY). toInt ();
	minimumSelect	-> setValue (min_loop_frequency / MHz (1));
	max_loop_frequency	=
	 s -> value ("max_loop_frequency", MAX_LOOP_FREQUENCY). toInt ();
	maximumSelect	-> setValue (max_loop_frequency / MHz (1));

	k	= s -> value ("spectrumAmplitudeSlider", 50). toInt ();
	spectrumAmplitudeSlider	-> setValue (k);

	k	= s -> value ("attenuationSlider", 50). toInt ();
	attenuationSlider	-> setValue (k);

	k	= s -> value ("IQbalanceSlider", 0). toInt ();
	IQbalanceSlider		-> setValue (k);	// will signal

	h	= s -> value ("inputModeSelect", "I and Q"). toString ();
	k	= inputModeSelect -> findText (h);
	if (k != -1) {
	   inputModeSelect -> setCurrentIndex (k);
	   setInputMode (h);
	}

	k	= s -> value ("VolumeSlider", 10). toInt ();
	VolumeSlider		-> setValue (k);	// will trigger

	h	= s -> value ("fmFilterSelect", "---"). toString ();
	k	= fmFilterSelect -> findText (h);
	if (k != -1) {
	   fmFilterSelect -> setCurrentIndex (k);
	   setfmBandwidth (h);
	}

	k	= s -> value ("fmFilterDegree", 21). toInt ();
	setfmBandwidth (k);
	fmFilterDegree		-> setValue (k);

	h	= s -> value ("fmMode", "stereo"). toString ();
	k	= fmMode -> findText (h);
	if (k != - 1) {
	   fmMode -> setCurrentIndex (k);
	   setfmMode (h);
	}

	h	= s -> value ("fmDecoder", "fm1decoder"). toString ();
	k	= fmDecoder -> findText (h);
	if (k != -1) {
	   fmDecoder -> setCurrentIndex (k);
	   setfmDecoder (h);
	}

	h	= s -> value ("fmRdsSelector", "rds1"). toString ();
	k	= fmRdsSelector -> findText (h);
	if (k != -1) {
	   fmRdsSelector -> setCurrentIndex (k);
	   setfmRdsSelector (h);
	}

	h	= s -> value ("fmView", "view 1"). toString ();
	k	= fmView -> findText (h);
	if (k != -1) {
	   fmView -> setCurrentIndex (k);
	   setfmView (h);
	}

	h	= s -> value ("fmChannelSelect", "stereo"). toString ();
	k	= fmChannelSelect -> findText (h);
	if (k != -1) {
	   fmChannelSelect -> setCurrentIndex (k);
	   setfmChannelSelector (h);
	}

	h	= s -> value ("fmDeemphasisSelector", "none"). toString ();
	k	= fmDeemphasisSelector -> findText (h);
	if (k != -1) {
	   fmDeemphasisSelector -> setCurrentIndex (k);
	   setfmDeemphasis (h);
	}

	k	= s -> value ("fmStereoSlider", 0). toInt ();
	fmStereoSlider	-> setValue (k);

	h	= s -> value ("fmLFcutoff", "off"). toString ();
	k	= fmLFcutoff -> findText (h);
	if (k != -1) {
	   fmLFcutoff -> setCurrentIndex (k);
	   setfmLFcutoff (h);
	}

	h	= s -> value ("logging", "log off"). toString ();
	k	= logging -> findText (h);
	if (k != -1) {
	   logging -> setCurrentIndex (k);
	   setLogging (h);
	}

	h	= s -> value ("deviceSelect", "default"). toString ();
	k	= deviceSelect	-> findText (h);
	if (k != -1) {
	   deviceSelect -> setCurrentIndex (k);
	   setDevice (h);
	}

	h	= s -> value ("streamOutSelector", "select output"). toString ();
	k	= streamOutSelector -> findText (h);
	if (k != -1) {
	   streamOutSelector -> setCurrentIndex (k);
	   setStreamOutSelector (k);
	}


//
//	we specify frequency in Khz
	k	= s -> value ("frequency", 89300). toInt ();
	Set_Tuner (Khz (k), &CurrentFreq);
	Display (CurrentFreq);

	if (IncrementIndex != 0)	// restart the incrementtimer if needed
	   autoIncrement_timer	-> start (IncrementInterval (IncrementIndex));
}

//
//	On start, we ensure that the streams are stopped so
//	that they can be restarted again.
void	RadioInterface::setStart	(void) {
bool	r = 0;

//	if (runMode == RUNNING)
//	   return;
//
//	always ensure that there is no datastream
	MyRig		-> stopReader ();
	our_audioSink 	-> stopWriter ();
//
//
	r = MyRig		-> restartReader ();

	qDebug ("Starting %d\n", r);
	if (!r) {
	   QMessageBox::warning (this, tr ("sdr"),
	                               tr ("Opening  input stream failed\n"));
	   return;
	}

	r = our_audioSink	-> restartWriter ();
	if (!r) {
	   QMessageBox::warning (this,
	                         tr ("sdr"),
	                         tr ("Opening outputstream failed\n"));
	   MyRig		-> stopReader ();
	   return;
	}

//	and finally: recall that starting overrules pausing
	pauseButton -> setText (QString ("Pause"));
	runMode	= RUNNING;
	Set_Tuner (CurrentFreq, &CurrentFreq);	// for the rtlsdr plugin
}

void	RadioInterface::TerminateProcess () {
	if (dumping) {
	   myFMprocessor	-> stopDumping ();
	   fclose (dumpFile);
	}

	if (ISettings != NULL)
	   dumpControlState	(ISettings);
	stop_lcdTimer		();
	stopIncrementing	();
//
//	The easiest way for a controlled going down is to
//	ensure that a "no rig" is the selected rig
	accept ();
	qDebug () <<  "Termination started";
}

void	RadioInterface::abortSystem (int d) {
	qDebug ("aborting for reason %d\n", d);
	accept ();
}
//

void	RadioInterface::setDevice (const QString & s) {
bool	success;
QString	file;

	if (dumping) {
	   myFMprocessor -> stopDumping ();
	   fclose (dumpFile);
	   dumping = false;
	   dumpButton	-> setText ("dump");
	}

	if (s == "dabstick") {
	   if (MyRig -> myIdentity () == STICK)
	      return;
	   MyRig	-> stopReader ();
	   delete myFMprocessor;
	   delete MyRig;
#ifndef __MINGW32__
	   MyRig	= new dabstick (dabRate, &success);
#else
	   MyRig	= new dabstick_dll (dabRate, &success);
#endif

	   if (!success) {
	      fprintf (stderr, "Cannot run with stick\n");
	      MyRig = new virtualInput ();
	   }	
	   else {
	      MyRig	-> setOffset (dongleOffset);
#ifndef __MINGW32__
	      ((dabstick *)MyRig)	-> freqCorrection (freqCorrectingValue);
#else
	      ((dabstick_dll *)MyRig)	-> freqCorrection (freqCorrectingValue);
#endif
	   }
	}
	else
	if (s == "file input") {
	   if (MyRig -> myIdentity () == FILEREADER)
	      return;
	   MyRig	-> stopReader ();
	   delete myFMprocessor;
	   delete MyRig;
	   file		= QFileDialog::getOpenFileName (this,
	                                                tr ("open file ..."),
	                                                QDir::homePath (),
	                                                tr ("raw data (*.raw)"));
	   file		= QDir::toNativeSeparators (file);
	   MyRig	= new fileReader (file. toAscii (). data (),
	                                  &success);
	   if (!success) {
	      delete MyRig;
	      MyRig = new virtualInput ();
	   }
	}
	else {	// s == "no device" 
	   MyRig	-> stopReader ();
	   delete myFMprocessor;
	   delete MyRig;
	   MyRig	= new virtualInput ();
	}

	myFMprocessor	= new fmProcessor (MyRig,
	                                    this,
	                                    our_audioSink,
	                                    dabRate,
	                                    fmRate,
	                                    workingRate,
	                                    audioRate,
	                                    displaySize,
	                                    spectrumSize,
	                                    averageCount,
	                                    repeatRate,
	                                    hfBuffer,
	                                    lfBuffer,
	                                    mySinCos);
	setAttenuation		(attenuationSlider	 -> value ());
	setVolume		(VolumeSlider		-> value ());
	setfmBandwidth		(fmFilterSelect		-> currentText ());
	setfmBandwidth		(fmFilterDegree		-> value ());
	setfmMode		(fmMode			-> currentText ());
	setfmRdsSelector	(fmRdsSelector		-> currentText ());
	setfmDecoder		(fmDecoder		-> currentText ());
	setfmView		(fmView			-> currentText ());
	setfmChannelSelector	(fmChannelSelect	-> currentText ());
	setfmDeemphasis		(fmDeemphasisSelector	-> currentText ());
	setfmStereoSlider	(fmStereoSlider		-> value ());
	setfmLFcutoff		(fmLFcutoff		-> currentText ());
	setLogging		(logging		-> currentText ());
}

void	RadioInterface::setHFAverager (void) {
	HFAverager	= !HFAverager;
	myFMprocessor	-> setFreezer (HFAverager);
}

void	RadioInterface::setLFAverager (void) {
	LFAverager	= !LFAverager;
	LFScope	-> setAverager (LFAverager);
}

void	RadioInterface::setAmplification (int a) {
	scopeAmplification	= 2 * a;
	LFScope -> setAmplification (a);
}

void	RadioInterface::setInputMode (const QString &s) {
	if (s == "I and Q")
	   inputMode	= IandQ;
	else
	if (s == "Q and I")
	   inputMode	= QandI;
	else
	if (s == "I Only")
	   inputMode	= I_Only;
	else
	if (s == "Q Only")
	   inputMode	= Q_Only;
	else
	   inputMode	= IandQ;
}

void	RadioInterface::setfmChannelSelector (const QString &s) {
	if (s == "stereo")
	   channelSelector	= fmProcessor::S_STEREO;
	else
	if (s == "Left")
	   channelSelector	= fmProcessor::S_LEFT;
	else
	if (s == "Right")
	   channelSelector	= fmProcessor::S_RIGHT;
	else
	if (s == "Left+Right")
	   channelSelector	= fmProcessor::S_LEFTplusRIGHT;
	else
	if (s == "Left-Right")
	   channelSelector	= fmProcessor::S_LEFTminusRIGHT;
	else		// the default
	   channelSelector	= fmProcessor::S_LEFT;
	myFMprocessor	-> setSoundMode	(channelSelector);
}

void	RadioInterface::setAttenuation (int n) {
int16_t	f	= IQBalanceDisplay	-> value ();
int16_t	bl, br;

	bl	= 100 - 2 * f;
	br	= 100 + 2 * f;
	attenuationLevelDisplay	-> display (n);
	currAttSliderValue	= 2 * n;
	attValueL	= currAttSliderValue * (float)bl / 100;
	attValueR	= currAttSliderValue * (float)br / 100;
	myFMprocessor	-> setAttenuation (attValueL, attValueR);
}
/*
 * 	the balance slider runs between -30 .. 30
 */
void	RadioInterface::setIQBalance (int n) {
int16_t	bl, br;

	IQBalanceDisplay -> display (n);
	bl		= 100 - 2 * n;
	br		= 100 + 2 * n;
	attValueL	= currAttSliderValue * (float)bl / 100;
	attValueR	= currAttSliderValue * (float)br / 100;
	myFMprocessor	-> setAttenuation (attValueL, attValueR);
}
/*
 * 	Handling the numeric keypad is boring, but inevitable
 */
void RadioInterface::addOne (void) {
	AddtoPanel (1);
}

void RadioInterface::addTwo (void) {
	AddtoPanel (2);
}

void RadioInterface::addThree (void) {
	AddtoPanel (3);
}

void RadioInterface::addFour (void) {
	AddtoPanel (4);
}

void RadioInterface::addFive (void) {
	AddtoPanel (5);
}

void RadioInterface::addSix (void) {
	AddtoPanel (6);
}

void RadioInterface::addSeven (void) {
	AddtoPanel (7);
}

void RadioInterface::addEight (void) {
	AddtoPanel (8);
}

void RadioInterface::addNine (void) {
	AddtoPanel (9);
}

void RadioInterface::addZero (void) {
	AddtoPanel (0);
}

void RadioInterface::addClear (void) {
	if (!lcd_timer -> isActive ())
	   return;

	stop_lcdTimer	();
//	we know that the increment timer is not running

	ClearPanel ();
	Display (CurrentFreq);
}
//
//	Note that in fm receiver, user commands are in KHz
//
void RadioInterface::AcceptFreqinKhz (void) {
int32_t	p;

	if (!lcd_timer -> isActive ())
	   return;

	stop_lcdTimer	();
//	we know that the increment timer is not running,
//	so there is no reason to stop it
	p = getPanel ();
	ClearPanel ();
	Set_Tuner (p * 1000, &CurrentFreq);
	Display (CurrentFreq);
}

void RadioInterface::addCorr (void) {
	if (!lcd_timer -> isActive ()) 
	   return;

	stop_lcdTimer	();
//	increment timer cannot run now so no reason
//	to stop it
	Panel = (Panel - (Panel % 10)) / 10;
	Display	(Panel);
	lcd_timer	-> start (5000);		// restart timer
}
//
//	In AddtoPanel, we might have any of the two
//	timers running, do not know which one though
void	RadioInterface::AddtoPanel (int16_t n) {
	stop_lcdTimer		();
	stopIncrementing	();

	Panel		= 10 * Panel + n;
	lcd_Frequency	-> setNumDigits (7);
	lcd_Frequency	-> display ((int)Panel);
	lcd_timer	-> start (5000);		// restart timer
}

void	RadioInterface::ClearPanel (void) {
	Panel = 0;
}

int	RadioInterface::getPanel (void) {
	return Panel;
}

void RadioInterface::decT5 (void) {
	IncrementFrequency (mapIncrement (-1));
}

void RadioInterface::decT50 (void) {
	IncrementFrequency (mapIncrement (-10));
}

void RadioInterface::decT500 (void) {
	IncrementFrequency (mapIncrement (-100));
}

void RadioInterface::decT5000 (void) {
	IncrementFrequency (mapIncrement (-1000));
}

void RadioInterface::incT5 (void) {
	IncrementFrequency (mapIncrement (1));
}

void RadioInterface::incT50 (void) {
	IncrementFrequency (mapIncrement (10));
}

void RadioInterface::incT500 (void) {
	IncrementFrequency (mapIncrement (100));
}

void RadioInterface::incT5000 (void) {
	IncrementFrequency (mapIncrement (1000));
}
//
//	Increment frequency: with amount N, depending
//	on the mode of operation
//
int32_t 	RadioInterface::mapIncrement (int32_t n) {
	return n * 1000;
}
//
//	IncrementFrequency is called from the handlers
//	for the autoincrement, the increment and the
//	++, + etc knobs.
//	Deviations within 15K of the VFO are handled with an
//	offset "onscreen"
void	RadioInterface::IncrementFrequency (int32_t n) {
int32_t	vfoFreq;
	stop_lcdTimer	();
	stopIncrementing	();

	MyRig	-> getVFOFrequency (&vfoFreq);
	if (abs (vfoFreq - (CurrentFreq + n)) > dabRate / 4)
	   Set_Tuner (CurrentFreq + n, &CurrentFreq);
	else {
	   CurrentFreq += n;
	   LOFrequency	= CurrentFreq - vfoFreq;
	   myFMprocessor	-> set_LocalOscillator (LOFrequency);
	}
	Display (CurrentFreq);
}
//	AdjustFrequency is called whenever someone clicks
//	with the button on the screen. The amount
//	has to be multiplied with 1000
void	RadioInterface::AdjustFrequency (int n) {
	IncrementFrequency (n * 1000);
}
//
//	Whenever the mousewheel is changed, the frequency
//	is adapted
void	RadioInterface::wheelEvent (QWheelEvent *e) {
	if (e -> delta () > 0)
	   IncrementFrequency (KHz (1));
	else
	   IncrementFrequency (-KHz (1));
}

void	RadioInterface::Zoom_in (int n) {
	LFScope	-> setZoompoint (n);
}

void	RadioInterface::Zoom_out (int n) {
	(void)n;
	LFScope -> resetZoompoint ();
}
//
//	Set_Tuner accepts freq in Hz.
//	It handles the whole range of frequencies,
//	including the 3-th harmonic trick for the pmsdr
//
//
//	The generic Set_Tuner. When it is too difficult
//	a specialized set_tuner is dispatched
void	RadioInterface::Set_Tuner (int32_t n, int32_t *c) {
int32_t	vfo;
int32_t	VFOfrequency;
int16_t	filter;
//
	VFOfrequency		= n;

	if (!MyRig -> legalFrequency (VFOfrequency)) {
	   Set_Tuner (MyRig -> defaultFrequency (), c);
	   return;
	}

	MyRig		-> setVFOFrequency	(VFOfrequency);
	MyRig		-> getVFOFrequency	(&vfo);
	vfoFrequency		= vfo;
	LOFrequency	= n - vfo;
	LFScope		-> setZero (0);
	LFScope		-> clearAverage	();
	myFMprocessor	-> resetRds	();
	*c		= n;
}
//
//===== code for auto increment/decrement
//	lots of code for something simple,

static inline
bool	frequencyInBounds (int32_t f, int32_t l, int32_t u) {
	return l <= f && f <= u;
}

int32_t	RadioInterface::IncrementInterval (int16_t index) {
	if (index < 0)
	   index = - index;

	if (index == 0)
	   index = 1;
	if (index >= delayTableSize)
	   index = delayTableSize;

	return 1000 * delayTable [index - 1];
}

void	RadioInterface::set_incrementFlag (int16_t incr) {
char temp [128];

	if (incr == 0) {
	   incrementingFlag ->
	                 setStyleSheet ("QLabel {background-color:blue}");
	   incrementingFlag -> setText (" ");
	   return;
	}
	if (incr < 0)
	   sprintf (temp, " << %d", IncrementInterval (incr) / 1000);
	else 
	   sprintf (temp, "%d >> ", IncrementInterval (incr) / 1000);
	incrementingFlag ->
	                 setStyleSheet ("QLabel {background-color:green}");
	incrementingFlag -> setText (temp);
}
//
void	RadioInterface::autoIncrement_timeout (void) {
int32_t	amount;
int32_t	frequency	= CurrentFreq;
int32_t	low, high;

	amount =  fmIncrement;

	if (IncrementIndex < 0)
	   amount = - amount;
//
//	Very stupid code, but remember that our user might have
//	changed the modus from am to fm or vv
	frequency	+= amount;
	low	= min_loop_frequency;
	high	= max_loop_frequency;
	                       
	if ((IncrementIndex < 0) &&
	   !frequencyInBounds (frequency, low, high))
	   frequency = max_loop_frequency;

	if ((IncrementIndex > 0) &&
	   !frequencyInBounds (frequency, low, high))
	   frequency = min_loop_frequency;

	Set_Tuner (frequency, &CurrentFreq);
	Display (CurrentFreq);
	autoIncrement_timer	-> start (IncrementInterval (IncrementIndex));
}
//
//	stopIncrementing is called from various places to
//	just interrupt the autoincrementing
void	RadioInterface::stopIncrementing (void) {
	set_incrementFlag (0);

	if (autoIncrement_timer	-> isActive ())
	   autoIncrement_timer -> stop ();

	IncrementIndex = 0;
}

void	RadioInterface::autoIncrementButton (void) {
	stop_lcdTimer ();

	if (autoIncrement_timer	-> isActive ())
	   autoIncrement_timer -> stop ();

	if (++IncrementIndex > delayTableSize)
	   IncrementIndex = delayTableSize;

	if (IncrementIndex == 0) {
	   set_incrementFlag (0);
	   return;
	}
//
	autoIncrement_timer	-> start (IncrementInterval (IncrementIndex));
	set_incrementFlag (IncrementIndex);
}

void	RadioInterface::autoDecrementButton (void) {
	stop_lcdTimer ();

	if (autoIncrement_timer	-> isActive ())
	   autoIncrement_timer -> stop ();

	if (--IncrementIndex < - delayTableSize)
	   IncrementIndex = - delayTableSize;

	if (IncrementIndex == 0) {
	   set_incrementFlag (0);
	   return;
	}
//
	autoIncrement_timer	-> start (IncrementInterval (IncrementIndex));
	set_incrementFlag (IncrementIndex);
}

void	RadioInterface::set_fm_increment (int v) {
	fmIncrement	= (int32_t) (v * 1000);
}

void	RadioInterface::set_freqCorrector (int v) {
	freqCorrectingValue	= v;
	if (MyRig -> myIdentity () != STICK)
	   return;
#ifndef	__MINGW32__
	((dabstick *)MyRig) -> freqCorrection (v);
#else
	((dabstick_dll *)MyRig) -> freqCorrection (v);
#endif

}

//
//	min and max frequencies are specified in Mhz

void	RadioInterface::set_minimum	(int f) {
	min_loop_frequency	= MHz (f);
}

void	RadioInterface::set_maximum	(int f) {
	max_loop_frequency	= MHz (f);
}

void	RadioInterface::IncrementButton (void) {
	stopIncrementing ();
	Set_Tuner (CurrentFreq + fmIncrement, &CurrentFreq);
	Display (CurrentFreq);
}

void	RadioInterface::DecrementButton (void) {
	stopIncrementing ();
	Set_Tuner (CurrentFreq - fmIncrement, &CurrentFreq);
	Display (CurrentFreq);
}
//
//=======================end of incrementing code===========================
void	RadioInterface::stop_lcdTimer (void) {
	if (lcd_timer -> isActive ())
	   lcd_timer -> stop ();
}

void	RadioInterface::updateTimeDisplay (void) {
QDateTime	currentTime = QDateTime::currentDateTime ();

	timeDisplay	-> setText (currentTime.
	                            toString (QString ("dd.MM.yy:hh:mm:ss")));

	if (myFMprocessor	-> ok ())
	   showStrength ();
}

void	RadioInterface::lcd_timeout (void) {
	Panel		= 0;			// throw away anything
	Display (CurrentFreq);
}

void	RadioInterface::key_timeout (void) {
	KeyhandlingMode	= NORMAL;
}

void	RadioInterface::set_dumping (void) {
	if (MyRig -> myIdentity () != STICK)
	   return;

	if (dumping) {
	   myFMprocessor	-> stopDumping ();
	   fclose (dumpFile);
	   dumping = false;
	   dumpButton	-> setText ("dump");
	   return;
	}

	QString file = QFileDialog::getSaveFileName (this,
	                                     tr ("open file ..."),
	                                     QDir::homePath (),
	                                     tr ("raw data (*.raw)"));
	file	= QDir::toNativeSeparators (file);
	dumpFile	= fopen (file. toAscii (). data (), "w");
	if (dumpFile == NULL) {
	   qDebug () << "cannot open " << file. toAscii (). data ();
	   return;
	}
	myFMprocessor	-> startDumping (dumpFile);
	dumpButton	-> setText ("writing");
	dumping		= true;
}

/*
 *	there is a tremendous amount of signal/slot connections
 *	The local connects, knobs, sliders and displays,
 *	are connected here.
 */
void	RadioInterface::localConnects (void) {
	connect (startButton, SIGNAL (clicked ()),
	              this, SLOT (setStart ()));
	connect (streamOutSelector, SIGNAL (activated (int)),
	              this, SLOT (setStreamOutSelector (int)));
	connect (QuitButton, SIGNAL (clicked ()),
	              this, SLOT (TerminateProcess (void)));
	connect (HFAverageButton, SIGNAL (clicked (void)),
	              this, SLOT (setHFAverager (void)));
	connect (LFAverageButton, SIGNAL (clicked (void)),
	              this, SLOT (setLFAverager (void)));
	connect (spectrumAmplitudeSlider, SIGNAL (valueChanged (int)),
	         this,  SLOT (setAmplification (int)));
	connect (deviceSelect, SIGNAL (activated (const QString &)),
	              this, SLOT (setDevice (const QString &)));

	connect (dumpButton, SIGNAL (clicked (void)),
	              this, SLOT (set_dumping (void)));

	connect (fc_plus, SIGNAL (clicked (void)),
	              this, SLOT (autoIncrementButton (void)));
	connect (fc_minus, SIGNAL (clicked (void)),
	              this, SLOT (autoDecrementButton (void)));
	connect (f_plus, SIGNAL (clicked (void)),
	              this, SLOT (IncrementButton (void)));
	connect (f_minus, SIGNAL (clicked (void)),
	              this, SLOT (DecrementButton (void)));
	connect (fm_increment, SIGNAL (valueChanged (int)),
	              this, SLOT (set_fm_increment (int)));

	connect (freqCorrector, SIGNAL (valueChanged (int)),
	              this, SLOT (set_freqCorrector (int)));
	connect (minimumSelect, SIGNAL (valueChanged (int)),
	              this, SLOT (set_minimum (int)));
	connect (maximumSelect, SIGNAL (valueChanged (int)),
	              this, SLOT (set_maximum (int)));

	connect (logging, SIGNAL (activated (const QString &)),
	              this, SLOT (setLogging (const QString &)));
	connect (logSaving, SIGNAL (clicked (void)),
	              this, SLOT (setLogsaving (void)));

	connect (attenuationSlider, SIGNAL (valueChanged (int) ),
		      this, SLOT (setAttenuation (int) ) );
	connect (IQbalanceSlider, SIGNAL (valueChanged (int) ),
		      this, SLOT (setIQBalance (int) ) );
	connect (HFplotterView, SIGNAL (activated (const QString &)),
	              this, SLOT (setHFplotterView (const QString &)));
/*
 *	Mode setters
 */
	connect (inputModeSelect, SIGNAL (activated(const QString&) ),
	              this, SLOT (setInputMode (const QString&) ) );
/*
 */
	connect (fmChannelSelect, SIGNAL (activated (const QString &)),
	               this, SLOT (setfmChannelSelector (const QString &)));
	connect	(pauseButton, SIGNAL (clicked (void)),
	               this, SLOT (clickPause (void)));
/*
 *	connections for the console, first the digits
 */
	connect (add_one,   SIGNAL (clicked() ), this, SLOT (addOne() ) );
	connect (add_two,   SIGNAL (clicked() ), this, SLOT (addTwo() ) );
	connect (add_three, SIGNAL (clicked() ), this, SLOT (addThree() ) );
	connect (add_four,  SIGNAL (clicked() ), this, SLOT (addFour() ) );
	connect (add_five,  SIGNAL (clicked() ), this, SLOT (addFive() ) );
	connect (add_six,   SIGNAL (clicked() ), this, SLOT (addSix() ) );
	connect (add_seven, SIGNAL (clicked() ), this, SLOT (addSeven() ) );
	connect (add_eight, SIGNAL (clicked() ), this, SLOT (addEight() ) );
	connect (add_nine,  SIGNAL (clicked() ), this, SLOT (addNine() ) );
	connect (add_zero,  SIGNAL (clicked() ), this, SLOT (addZero() ) );
/*
 *	function buttons
 */
	connect (khzSelector, SIGNAL (clicked ()),
	                      this, SLOT (AcceptFreqinKhz ()));
	connect (add_correct, SIGNAL (clicked() ), this, SLOT (addCorr() ) );
	connect (add_clear,   SIGNAL (clicked() ), this, SLOT (addClear() ) );
/*
 *	The increment and decrement buttons
 */
	connect (dec_5,	   SIGNAL (clicked() ), this, SLOT (decT5() ) );
	connect (dec_50,   SIGNAL (clicked() ), this, SLOT (decT50() ) );
	connect (dec_500,  SIGNAL (clicked() ), this, SLOT (decT500() ) );
	connect (inc_5,	   SIGNAL (clicked() ), this, SLOT (incT5() ) );
	connect (inc_50,   SIGNAL (clicked() ), this, SLOT (incT50() ) );
	connect (inc_500,  SIGNAL (clicked() ), this, SLOT (incT500() ) );
/*
 */
	connect (fmFilterSelect, SIGNAL (activated (const QString &)),
	         this, SLOT (setfmBandwidth (const QString &)));
	connect (fmFilterDegree, SIGNAL (valueChanged (int)),
	         this, SLOT (setfmBandwidth (int)));
	connect (fmMode, SIGNAL (activated (const QString &)),
	               this, SLOT (setfmMode (const QString &)));
	connect (fmRdsSelector, SIGNAL (activated (const QString &)),
	               this, SLOT (setfmRdsSelector (const QString &)));
	connect (fmDecoder, SIGNAL (activated (const QString &)),
	               this, SLOT (setfmDecoder (const QString &)));
	connect (fmView, SIGNAL (activated (const QString &)),
	               this, SLOT (setfmView (const QString &)));
	connect	(VolumeSlider, SIGNAL (valueChanged (int)),
	              this, SLOT (setVolume (int)));
	connect (fmStereoSlider, SIGNAL (valueChanged (int)),
	              this, SLOT (setfmStereoSlider (int)));
	connect (fmDeemphasisSelector, SIGNAL (activated (const QString &)),
	              this, SLOT (setfmDeemphasis (const QString &)));
	connect (fmLFcutoff, SIGNAL (activated (const QString &)),
	              this, SLOT (setfmLFcutoff (const QString &)));
}

void	RadioInterface::setfmStereoSlider	(int n) {
	myFMprocessor	-> setSoundBalance (n);
	balanceDisplay	-> display (n);
}
//	Deemphasis	= 50 usec (3183 Hz, Europe)
//	Deemphasis	= 75 usec (2122 Hz US)
void	RadioInterface::setfmDeemphasis	(const QString& s) {
	if (s == "50")
	   myFMprocessor	-> setDeemphasis (50);
	else
	if (s == "75")
	   myFMprocessor	-> setDeemphasis (75);
	else
	   myFMprocessor	-> setDeemphasis (1);
}

void	RadioInterface::setVolume (int v) {
	Volume	= (int16_t)v;
	myFMprocessor	-> setVolume (Volume);
}
//

void	RadioInterface::setCRCErrors	(int n) {
	crcErrors	-> display (n);
}

void	RadioInterface::setSyncErrors	(int n) {
	syncErrors	-> display (n);
}

void	RadioInterface::setbitErrorRate	(double v) {
	bitErrorRate	-> display (v);
}

void	RadioInterface::setGroup	(int n) {
	(void)n;
//	rdsGroupDisplay	-> display (n);
}

void	RadioInterface::setPTYCode	(int n) {
	(void)n;
//	rdsPTYDisplay	-> display (n);
}

void	RadioInterface::setAFDisplay	(int n) {
	rdsAFDisplay	-> display (n);
}

void	RadioInterface::setPiCode	(int n) {
int32_t	t	= CurrentFreq;

	if ((frequencyforPICode != t) || (n != 0)) {
	   currentPIcode	= n;
	   frequencyforPICode = t;
	}

	rdsPiDisplay	-> display (n);
}

void	RadioInterface::clearStationLabel (void) {
	StationLabel = QString ("");
	stationLabelTextBox	-> setText (StationLabel);
}
//
//	Note: although s is a char * type, its value does not
//	end with a zero, so it is not a C string
void	RadioInterface::setStationLabel (char *s, int size) {
uint16_t	i;

	StationLabel = QString ("");
	for (i = 0; i < size; i ++)
	   StationLabel. append (QString (QChar (s [i])));
	stationLabelTextBox	-> setText (StationLabel);
}

void	RadioInterface::setMusicSpeechFlag (int n) {
	if (n != 0)
	   speechLabel -> setText (QString ("music"));
	else
	   speechLabel -> setText (QString ("speech"));
}

void	RadioInterface::clearMusicSpeechFlag (void) {
	speechLabel	-> setText (QString (""));
}

void	RadioInterface::clearRadioText (void) {
	RadioText = QString ("");
	radioTextBox	-> setText (RadioText);
}
//
//	Note that, although s is a "char *", it is not a C string,
//	(no zero at the end)
void	RadioInterface::setRadioText (char *s, int size) {
uint16_t	i;
	RadioText	= QString ("");
	for (i = 0; i < size; i ++) {
	   RadioText. append (QString (QChar (s [i])));
	}
	radioTextBox	-> setText (RadioText);
}

void	RadioInterface::setRDSisSynchronized (bool syn) {
	if (!syn)
	   rdsSyncLabel -> setStyleSheet ("QLabel {background-color:red}");
	else
	   rdsSyncLabel -> setStyleSheet ("QLabel {background-color:green}");
}
//

void	RadioInterface::setfmMode (const QString &s) {
	myFMprocessor	-> setfmMode (s == "stereo");
}

void	RadioInterface::setfmRdsSelector (const QString &s) {
	rdsModus = (s == "rds 1" ? rdsDecoder::RDS1 :
	            s == "rds 2" ? rdsDecoder::RDS2 : 
	            rdsDecoder::NO_RDS);
	myFMprocessor	-> setfmRdsSelector (rdsModus);
}

void	RadioInterface::setfmView (const QString &s) {
int16_t	d = 1;

	if (s == "view 1")
	   d = 1;
	if (s == "view 2")
	   d = 2;
	if (s == "view 3")
	   d = 3;
	if (s == "view 4")
	   d = 4;
	if (s == "view 5")
	   d = 5;
	if (s == "view 6")
	   d = 6;
	if (s == "view 7")
	   d = 7;
	if (s == "view 8")
	   d = 8;

	viewKeuze	= d;
	myFMprocessor	-> setView (d);
}

void	RadioInterface::setfmDecoder (const QString &s) {
int8_t	decoder	= 0;

	if (s == "fm1decoder")
	   decoder = fm_Demodulator::FM1DECODER;
	else
	if (s == "fm2decoder")
	   decoder = fm_Demodulator::FM2DECODER;
	else
	if (s == "fm3decoder")
	   decoder = fm_Demodulator::FM3DECODER;
	else
	if (s == "fm4decoder")
	   decoder = fm_Demodulator::FM4DECODER;
	else
	   decoder = fm_Demodulator::FM5DECODER;

	myFMprocessor	-> setFMdecoder (decoder);
	decoderDisplay	-> setText (QString (myFMprocessor -> nameofDecoder ()));
}
//
void	RadioInterface::setfmLFcutoff (const QString &s) {
	if (s == "off")
	   myFMprocessor	-> setLFcutoff (-1);
	else
	   myFMprocessor	-> setLFcutoff ((int32_t)(s. toInt ()));
}
//
static inline
int32_t numberofDigits (int32_t f) {

	if (f < 100000)
	   return 6;
	if (f < 100000000)
	   return 8;
	if (f < 1000000000)
	   return 9;
	return 10;
}

void	RadioInterface::Display (int32_t freq) {
int32_t nd	= numberofDigits (freq);
 
	lcd_Frequency	-> setNumDigits (6);
	lcd_Frequency	-> display ((int)freq / 1000);
}

void	RadioInterface::setfmBandwidth (const QString &s) {

	if (s == "---") {
	   fmBandwidth	= 0.95 * fmRate;
	   return;
	}

	fmBandwidth = Khz (s. toInt ());
	myFMprocessor	-> setBandwidth (fmBandwidth / 2);
}

void	RadioInterface::setfmBandwidth (int32_t b) {
	myFMprocessor	-> setBandfilterDegree	(b);
}

//
void	RadioInterface::showStrength (void) {
	pilotStrength	->
	           display ((int)(myFMprocessor -> getPilotStrength ()));
	noiseStrength	->
	           display ((int)(myFMprocessor -> getNoiseStrength ()));
	rdsStrength	->
	           display ((int)(myFMprocessor -> getRdsStrength ()));
	if (!myFMprocessor -> isLocked ())
	   pll_isLocked -> setStyleSheet ("QLabel {background-color:red}");
	else
	   pll_isLocked -> setStyleSheet ("QLabel {background-color:green}");
	dc_component	-> display (myFMprocessor -> get_DcComponent ());
}

void	RadioInterface::setLogging (const QString &s) {

	if (s == "log off") {
	   if (logTime != 0)
	      stopLogging	();
	   return;
	}

	if (log_timer	-> isActive ())
	   log_timer -> stop ();

	if (s == "log 1 sec")
	   startLogging (1);
	else
	if (s == "log 2 sec")
	   startLogging (2);
	else
	if (s == "log 3 sec")
	   startLogging (3);
	else
	if (s == "log 4 sec")
	   startLogging (4);
	else
	   startLogging (5);
}

void	RadioInterface::startLogging (int16_t t) {

	if (log_timer	-> isActive ())
	   log_timer -> stop ();
	logTime		= t;
	log_timer	-> start (t * 1000);
}

void	RadioInterface::stopLogging	(void) {
	if (logTime == 0)	// just return, logging was not active
	   return;

	logTime		= 0;
	if (log_timer	-> isActive ())
	   log_timer -> stop ();
}

void	RadioInterface::setLogsaving (void) {

	if (logFile != NULL) {		// just stop it
	   fclose (logFile);
	   logFile		= NULL;
	   logSaving		-> setText ("save");
	   return;
	}
	else {
           QString file	= QFileDialog::getSaveFileName (this,
	                                        tr ("open file .."),
	                                        QDir::homePath (),
	                                        tr ("text (*.txt)"));
	   logFile		= fopen (file.toAscii (). data (), "w");
	   if (logFile == NULL) {
	      QMessageBox::warning (this, tr ("sdr"),
	                                  tr ("/dev/null is used"));
	
//	      logFile	= NULL;		// remains zero
	   }
	   else {	// logFile != NULL
	      logSaving		-> setText ("halt");
	      fprintf (logFile, "\nlogging starting\n\n\n");
	   }
	}
}

void	RadioInterface::log_timeout (void) {
QDateTime	currentTime = QDateTime::currentDateTime ();
DSPFLOAT	signal, pilot, rds;
char	s [1024];

	signal		= myFMprocessor -> getNoiseStrength	();
	pilot		= myFMprocessor	-> getPilotStrength	();
	rds		= myFMprocessor	-> getRdsStrength	();

	sprintf (s,
        "%s : Freq = %d,\n PI code = %4X, noise = %2d, pilot = %2d,  rds = %2d %s\n",
	currentTime.toString (QString ("dd.MM.yy:hh:mm:ss")). toStdString (). c_str (),
	(int)CurrentFreq,
	currentPIcode,
	(int)signal,
	(int)pilot,
	(int)rds,
	myFMprocessor -> isLocked () ? "pilot" : "no pilot");

	fputs (s, stderr);
//	and into the logfile
	if (logFile != NULL)
	   fputs (s, logFile);

	log_timer	-> start (logTime * 1000);
}

void	RadioInterface::clickPause (void) {

	if (runMode == IDLE)
	   return;

	if (runMode == RUNNING) {
	   if (autoIncrement_timer	-> isActive ())
	      autoIncrement_timer 	-> stop ();

	   MyRig		-> stopReader	();
	   our_audioSink	-> stopWriter		();
	   pauseButton -> setText (QString ("Continue"));
	   runMode = PAUSED;
	}
	else
	if (runMode == PAUSED) {
	   if (IncrementIndex != 0)	// restart the incrementtimer if needed
	      autoIncrement_timer	-> start (IncrementInterval (IncrementIndex));
	   MyRig		-> restartReader ();
	   our_audioSink	-> restartWriter ();
	   pauseButton -> setText (QString ("Pause"));
	   runMode = RUNNING;
	}
}

//
//
//	OK, here we map the "keystrokes" onto commands
//	We have some "direct keys" for some main functions,
//	and further we have "Ctrl" keys for the many other functions
//
void	RadioInterface::getKeyStroke (int key) {
QInputDialog	mydialog;
int32_t	f;

	if (key != 0)
	   fprintf (stderr, "Key %x touched\n", key);

	if (key == Qt::Key_Q) {
	   TerminateProcess ();
	   return;
	}

	if (key == Qt::Key_R) {
	   KeyhandlingMode	= NORMAL;
	   return;
	}

	if (key == Qt::Key_S) {
	   KeyhandlingMode 	= NORMAL;
	   setStart ();
	   return;
	}

	if (key_timer -> isActive ())
	   key_timer -> stop ();

	switch (KeyhandlingMode) {
	   case NORMAL: 
	      switch (key) {
	         case Qt::Key_Control:
	            KeyhandlingMode	= CONTROL;
	            key_timer		-> start (5000);
	            return;

	         case Qt::Key_Alt:
	            KeyhandlingMode	= ALT;
	            key_timer		-> start (5000);
	            return;

	         default:
	            return;
	      }

	   case CONTROL: 	// set frequencies
	      switch (key) {
	         case Qt::Key_Control:
	            KeyhandlingMode	= NORMAL;
	            return;

	         case Qt::Key_Up:
	            IncrementFrequency (Khz (1));
	            KeyhandlingMode	= NORMAL;
	            return;

	         case Qt::Key_Down:
	            IncrementFrequency (Khz (-1));
	            KeyhandlingMode	= NORMAL;
	            return;

	         case Qt::Key_PageUp:
	            IncrementFrequency (Khz (100));
	            KeyhandlingMode	= NORMAL;
	            return;

	         case Qt::Key_PageDown:
	            IncrementFrequency (Khz (-100));
	            KeyhandlingMode	= NORMAL;
	            return;

	         case Qt::Key_Minus:
	            IncrementFrequency (mapIncrement (-1));
	            KeyhandlingMode	= NORMAL;
	            return;

	         case Qt::Key_Plus:
	            IncrementFrequency (mapIncrement (1));
	            KeyhandlingMode	= NORMAL;
	            return;

	         case Qt::Key_U:
	            IncrementFrequency (fmRate);
	            KeyhandlingMode	= NORMAL;
	            return;

	         case Qt::Key_D:
	            IncrementFrequency (-fmRate);
	            KeyhandlingMode	= NORMAL;
	            return;

	         case Qt::Key_A:
	             f	=
	                  mydialog. getInt (this, tr("new Frequency"),
	                                            tr("Freq"));
	             f	= f * KHz (1);
	             Set_Tuner (f, &CurrentFreq);
	             Display (f);
	             KeyhandlingMode	= NORMAL;
	             return;
	             
	         default:
	            return;
	      } 
	   
	   case ALT:
	      switch (key) {
	         case Qt::Key_A:
	            KeyhandlingMode	= NORMAL;
	            return;

	         case Qt::Key_F:
	            KeyhandlingMode	= NORMAL;
	            return;
	
	         case Qt::Key_T:
	            KeyhandlingMode	= NORMAL;
	            return;

	         default:
	            KeyhandlingMode	= NORMAL;
	            return;
	       }
	      return;
	}
}
//
//	For the HF scope, we only do the displaying (and X axis)
//	in the GUI environment. The FM processor prepares "views"
//	and punt these views into a shared buffer. If the buffer is
//	full, a signal is sent.
void	RadioInterface::hfBufferLoaded (int amount) {
double	*X_axis		= (double *)alloca (displaySize * sizeof (double));
double	*Y_values	= (double *)alloca (displaySize * sizeof (double));
int16_t	i, j;
double	temp	= (double)dabRate / 2 / displaySize;

//	first X axis labels
	for (i = 0; i < displaySize; i ++)
	   X_axis [i] = 
	      ((double)vfoFrequency - (double)(dabRate / 2) +
	        (double)((i) * (double) 2 * temp)) / ((double)1000);
//
//	get the buffer data
	hfBuffer -> getDataFromBuffer (Y_values, displaySize);
	HFScope -> Display (X_axis, Y_values,
	                    scopeAmplification,
	                    (vfoFrequency + LOFrequency) / 1000);
}
//
//	The data is put into a shared buffer by the FM processor
//	a signal is given when a reasonable amount of data
//	is in this buffer
void	RadioInterface::lfBufferLoaded (int s) {
DSPCOMPLEX	*v = (DSPCOMPLEX *)alloca (s * sizeof (DSPCOMPLEX));
int32_t	i;

	lfBuffer -> getDataFromBuffer (v, s);
	for (i = 0; i < s; i ++)
	   LFScope -> addElement (v [i]);
}

//	do not forget that icnt and ocnt start with 1, due
//	to Qt list conventions
bool	RadioInterface::setupSoundOut (QComboBox	*streamOutSelector,
	                               paWriter		*our_audioSink,
	                               int32_t		cardRate,
	                               int16_t		*table) {
uint16_t	ocnt	= 1;
uint16_t	i;

	for (i = 0; i < our_audioSink -> numberofDevices (); i ++) {
	   const char *so =
	             our_audioSink -> outputChannelwithRate (i, cardRate);
	   qDebug ("Investigating Device %d\n", i);

	   if (so != NULL) {
	      streamOutSelector -> insertItem (ocnt, so, QVariant (i));
	      table [ocnt] = i;
	      qDebug (" (output):item %d wordt stream %d\n", ocnt , i);
	      ocnt ++;
	   }
	}

	qDebug () << "added items to combobox";
	return ocnt > 1;
}

void	RadioInterface::setStreamOutSelector (int idx) {
	if (idx == 0)
	   return;

	outputDevice = outTable [idx];
	if (!our_audioSink -> isValidDevice (outputDevice)) 
	   return;

	our_audioSink	-> stopWriter	();
	if (!our_audioSink -> selectDevice (outputDevice)) {
	   QMessageBox::warning (this, tr ("sdr"),
	                               tr ("Selecting  output stream failed\n"));
	   return;
	}

	qWarning () << "selected output device " << idx << outputDevice;
}

void	RadioInterface::setHFplotterView (const QString &s) {
	if (s == "waterfall") {
	   HFviewMode = WATERFALL_MODE;
	}
	else {	// s == "spectrumView"
	   HFviewMode = SPECTRUM_MODE;
	}

	HFScope	-> SelectView (HFviewMode);
}

