#
/*
 *
 *    Copyright (C)  2010, 2011, 2012
 *    Jan van Katwijk (J.vanKatwijk@gmail.com)
 *    JFF Consultancy
 *
 *    This file is part of the JSDR.
 *    Many of the ideas as implemented in JSDR are derived from
 *    other work, made available through the GNU general Public License. 
 *    All copyrights of the original authors are recognized.
 *
 *    JSDR is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    JSDR is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with JSDR; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef _GUI
#define _GUI

#include	"fm-constants.h"
#include	<QDialog>
#include	"ui_sdrgui.h"
#include	<qwt.h>
#include	<qwt_slider.h>
#include	<qwt_plot.h>
#include	<qwt_plot_curve.h>
#include	<qwt_plot_marker.h>
#include	<qwt_plot_grid.h>
#include	<qwt_dial.h>
#include	<qwt_dial_needle.h>
#include	<qwt_plot_spectrogram.h>
#include	<QTimer>
#include	<QQueue>
#include	<QWheelEvent>
#include	"scope.h"
#include	"iqdisplay.h"
#include	"virtual_input.h"
#include	"sincos.h"
#include	"fir-filters.h"
#include	"keyboardfilter.h"
#include	<QKeyEvent>
#include	"ringbuffer.h"
#include	"scope.h"
#include	"fft.h"

class	QSettings;
class	fmProcessor;
class	rdsDecoder;
class	downDecimator;
class	fft_scope;
class	Oscillator;
class	paWriter;

#define	HFSPECTRUM	0200
#define	LFSPECTRUM	0201

/*
 *	GThe main gui object. It inherits from
 *	QDialog and the generated form
 */
class RadioInterface: public QDialog,
		      private Ui_elektorSDR {
Q_OBJECT
public:
		RadioInterface		(QSettings	*,
	                                 int32_t,		// output rate
	                                 KeyboardFilter	*,	//
	                                 QWidget *parent = NULL);
		~RadioInterface		();

private:
	enum Keyboard {
	   NORMAL	= 0,
	   CONTROL	= 1,
	   ALT		= 2,
	   SHIFT	= 3
	};

	int8_t		KeyhandlingMode;
	KeyboardFilter	*filter;
	bool		doInit;
	int16_t		outputDevice;
	void		localConnects		(void);
	void		setInitialControlState	(void);
	
	void		dumpControlState	(QSettings *);
	void		setControlState		(QSettings *);

	RingBuffer<double>	*hfBuffer;
	RingBuffer<DSPCOMPLEX>	*lfBuffer;
	Scope		*HFScope;
	int16_t		scopeAmplification;
	fft_scope	*LFScope;
	bool		HFAverager;
	bool		LFAverager;

	QSettings	*ISettings;
	int32_t		dabRate;
	int32_t		fmRate;
	int32_t		workingRate;
	int32_t		audioRate;
	int16_t		displaySize;
	int16_t		rasterSize;
	int16_t		spectrumSize;
	int32_t		averageCount;
	int16_t		repeatRate;
	int16_t		fmRateMultiplier;
	int16_t		dabRateMultiplier;
	double		*displayBuffer;

	paWriter	*our_audioSink;
	int16_t		Volume;
	int8_t		channelSelector;
	SinCos		*mySinCos;
	virtualInput	*MyRig;
	int16_t		*outTable;
	int16_t		numberofDevices;

	uint8_t		HFviewMode;
	uint8_t		inputMode;
	int16_t		currAttSliderValue;
	DSPFLOAT	attValueL;
	DSPFLOAT	attValueR;

	int32_t		fmBandwidth;
	Oscillator	*LocalOscillator;
	int32_t		LOFrequency;
	int32_t		CurrentFreq;
	LowPassFIR	*fmFilter;

	int32_t		dongleOffset;
	int32_t		min_loop_frequency;
	int32_t		max_loop_frequency;
	void		ClearPanel		();
	void		AddtoPanel		(int16_t);
	int		getPanel		();
	void		CorrectPanel		();

	void		setDetectorScreen	(int16_t);

	DSPFLOAT	omega;

	int32_t		mapIncrement		(int32_t);
	int32_t		IncrementInterval	(int16_t);
	void		Set_Tuner		(int32_t, int32_t *);
	void		Display			(int32_t);
	QTimer		*lcd_timer;
	QTimer		*key_timer;
	QTimer		*autoIncrement_timer;
	int16_t		IncrementIndex;
	int32_t		autoIncrement_amount;
	int32_t		fmIncrement;
	
	void		set_incrementFlag	(int16_t);
	void		stopIncrementing	(void);
	int32_t		get_Increment_for	(int16_t);

	void		stop_lcdTimer		(void);
	int32_t		Panel;
	int16_t		CurrentRig;
	QTimer		*display_timer;
/*
 *	dumping
 */
	bool		dumping;
	FILE		*dumpFile;

	fmProcessor	*myFMprocessor;
	rdsDecoder	*myRdsDecoder;
	downDecimator	*audioDecimator;
	downDecimator	*rdsDecimator;
	int8_t		rdsModus;
	int8_t		viewKeuze;

	void		HandleSamples		(DSPCOMPLEX *, int32_t);
	QString		RadioText;
	QString		StationLabel;
	void		IncrementFrequency	(int32_t);

	DSPFLOAT	currentPilotStrength;
	DSPFLOAT	currentSignalStrength;
	int32_t		currentPIcode;
	int32_t		frequencyforPICode;
	int16_t		logTime;
	void		startLogging		(int16_t);
	QTimer		*log_timer;
	void		stopLogging		(void);
	FILE		*logFile;

	int8_t		runMode;
	void		showStrength		(void);
	int32_t		vfoFrequency;
	int32_t		freqCorrectingValue;
/*
 *	The private slots link the GUI elements
 *	to the GUI code
 */
private slots:
	void	setStart		(void);
	void	lcd_timeout		(void);
	void	key_timeout		(void);
	void	updateTimeDisplay	(void);
	void	clickPause		(void);

	void	setInputMode		(const QString &);
	void	addOne			(void);
	void	addTwo			(void);
	void	addThree		(void);
	void	addFour			(void);
	void	addFive			(void);
	void	addSix			(void);
	void	addSeven		(void);
	void	addEight		(void);
	void	addNine			(void);
	void	addZero			(void);
	void	addClear		(void);
	void	AcceptFreqinKhz		(void);
	void	addCorr			(void);

	void	setAttenuation		(int);
	void	setIQBalance		(int);

	void	setHFplotterView	(const QString &);
	void	setHFAverager		(void);
	void	setLFAverager		(void);

	void	setAmplification	(int);
	void	decT5			(void);
	void	decT50			(void);
	void	decT500			(void);
	void	decT5000		(void);
	void	incT5			(void);
	void	incT50			(void);
	void	incT500			(void);
	void	incT5000		(void);

	void	setStreamOutSelector	(int);
	void	abortSystem		(int);
	void	TerminateProcess	(void);
	void	setDevice		(const QString &);
	void	setVolume		(int);
	void	set_dumping		(void);

	void	setfmBandwidth		(const QString &);
	void	setfmBandwidth		(int);
	void	setfmMode		(const QString &);
	void	setfmRdsSelector	(const QString &);
	void	setfmView		(const QString &);
	void	setfmDecoder		(const QString &);
	void	setfmChannelSelector	(const QString &);
	void	setfmDeemphasis		(const QString &);
	void	setfmStereoSlider	(int);
	void	setfmLFcutoff		(const QString &);

	void	autoIncrement_timeout	(void);
	void	autoIncrementButton	(void);
	void	autoDecrementButton	(void);
	void	set_fm_increment	(int);
	void	set_freqCorrector	(int);
	void	set_minimum		(int);
	void	set_maximum		(int);
	void	IncrementButton		(void);
	void	DecrementButton		(void);

	bool	setupSoundOut		(QComboBox *, paWriter *,
	                                 int32_t, int16_t *);
public slots:
	void	hfBufferLoaded		(int);
	void	lfBufferLoaded		(int);
	void	getKeyStroke		(int);
	void	wheelEvent		(QWheelEvent *);
	void	setLogging		(const QString &);
	void	setLogsaving		(void);
	void	log_timeout		(void);
	void	AdjustFrequency		(int);
	void	Zoom_in			(int);
	void	Zoom_out		(int);
	void	setCRCErrors		(int);
	void	setSyncErrors		(int);
	void	setbitErrorRate		(double);
	void	setGroup		(int);
	void	setPTYCode		(int);
	void	setPiCode		(int);
	void	clearStationLabel	(void);
	void	setStationLabel		(char *, int);
	void	clearRadioText		(void);
	void	setRadioText		(char *, int);
	void	setAFDisplay		(int);
	void	setRDSisSynchronized	(bool);
	void	setMusicSpeechFlag	(int);
	void	clearMusicSpeechFlag	(void);
};

#endif

