#
/*
 *
 *    Copyright (C)  2010, 2011, 2012
 *    Jan van Katwijk (J.vanKatwijk@gmail.com)
 *    JFF Consultancy
 *
 *    This file is part of the JFF SDR (JSDR).
 *    Many of the ideas as implemented in JSDR are derived from
 *    other work, made available through the GNU general Public License. 
 *    All copyrights of the original authors are recognized.
 *
 *    JSDR is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    JSDR is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with JSDR; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include	"scope.h"

/*
 *	The "scope" combines the Qwt widgets and control for both
 *	the spectrumdisplay and the waterfall display. It is used
 *	two times: for the large front end display and for the
 *	display used by the detectors
 */
	Scope::Scope (QwtPlot		*scope,
	              uint16_t		displaysize,
	              uint16_t		rastersize) {
	Plotter			= scope;
	Displaysize		= displaysize;
	Rastersize		= rastersize;

	Waterfall		= new WaterfallViewer (Plotter,
	                                               Displaysize,
	                                               Rastersize);
	Spectrum		= NULL;
	CurrentWidget		= WATERFALL_MODE;
	connect (Waterfall,
	         SIGNAL (leftClicked (int)),
	         this,
	         SLOT (leftClicked (int)));
}

	Scope::~Scope (void) {
	if (Spectrum != NULL)
	   delete Spectrum;
	if (Waterfall != NULL)
	   delete Waterfall;
}

void	Scope::leftClicked (int n) {
	clickedwithLeft (n);
}

void	Scope::rightClicked (int n) {
	clickedwithRight (n);
}

void	Scope::SelectView (uint8_t n) {
	if (CurrentWidget == n)
	   return;

	if (n == SPECTRUM_MODE) {
	   disconnect (Waterfall,
	               SIGNAL (leftClicked (int)),
	               this,
	               SLOT (leftClicked (int)));
	   delete Waterfall;
	   Plotter	-> clear ();
	   Waterfall	= NULL;
	   Spectrum	= new SpectrumViewer  (Plotter,
	                                       Displaysize);
	   connect (Spectrum,
	            SIGNAL (leftClicked (int)),
	            this,
	            SLOT (leftClicked (int)));
	   connect (Spectrum,
	            SIGNAL (rightClicked (int)),
	            this,
	            SLOT (rightClicked (int)));
	   CurrentWidget = SPECTRUM_MODE;
	}
	else	// n == WATERFALL_MODE
	if (n == WATERFALL_MODE) {
	   disconnect (Spectrum,
	               SIGNAL (leftClicked (int)),
	               this,
	               SLOT (leftClicked (int)));
	   disconnect (Spectrum,
	               SIGNAL (rightClicked (int)),
	               this,
	               SLOT (rightClicked (int)));
	   delete Spectrum;
	   Plotter	-> clear ();
	   Spectrum	= NULL;
	   Waterfall	= new WaterfallViewer (Plotter,
	                                       Displaysize,
	                                       Rastersize);
	   connect (Waterfall,
	            SIGNAL (leftClicked (int)),
	            this,
	            SLOT (leftClicked (int)));
	   CurrentWidget = WATERFALL_MODE;
	}
}

void	Scope::Display (double	*x_axis,
	                double	*buffer,
	                double	amp,
	                int32_t	marker) {
	if (CurrentWidget == WATERFALL_MODE)
	   Waterfall	-> ViewWaterfall (x_axis,
	                                  buffer,
	                                  amp,
	                                  marker);
	else
	   Spectrum	-> ViewSpectrum (x_axis,
	                                 buffer,
	                                 amp,
	                                 marker);
}
/*
 *	The waterfall display
 */
	WaterfallViewer::WaterfallViewer (QwtPlot	*plot,
	                                  uint16_t	displaysize,
	                                  uint16_t	rastersize):
	                                  QwtPlotSpectrogram () {
QwtLinearColorMap colorMap (Qt::darkCyan, Qt::red);
QwtScaleWidget *rightAxis;
int	i, j;

	plot	-> clear ();
	plotgrid		= plot;
	this	-> Displaysize	= displaysize;
	this	-> Rastersize	= rastersize;
	colorMap. addColorStop	(0.1, Qt::cyan);
	colorMap. addColorStop	(0.4, Qt::green);
	colorMap. addColorStop	(0.7, Qt::yellow);
	this -> setColorMap (colorMap);
	OneofTwo		= 0;
	rightAxis = plotgrid -> axisWidget (QwtPlot::yRight);
// A color bar on the right axis
	rightAxis -> setColorBarEnabled (true);

	plotData = new double [Displaysize * Rastersize];
	for (i = 0; i < Rastersize; i ++)
 	   for (j = 0; j < Displaysize; j ++)
	      plotData [i * Displaysize + j] = 0.5;

	WaterfallData	= new SpectrogramData (plotData,
	                                       10000,
	                                       1000,
	                                       Rastersize,
	                                       Displaysize,
	                                       50.0);
	this -> setData (*WaterfallData);
	this -> setDisplayMode (QwtPlotSpectrogram::ImageMode, TRUE);
	rightAxis -> setColorMap (this -> data (). range(),
	                          this -> colorMap ());

	plotgrid	-> setAxisScale (QwtPlot::yRight,
	                                 this -> data (). range (). minValue (),
                                         this -> data (). range (). maxValue ()
	                                );

	plotgrid	-> enableAxis (QwtPlot::yRight);
	plotgrid	-> setAxisScale (QwtPlot::xBottom,
	                                 10000,
	                                 11000);
	plotgrid	-> enableAxis (QwtPlot::xBottom);
	plotgrid	-> enableAxis (QwtPlot::yLeft);
	plotgrid	-> setAxisScale (QwtPlot::yLeft, 0, Rastersize);

	Marker		= new QwtPlotMarker ();
	Marker		-> setLineStyle (QwtPlotMarker::VLine);
	Marker		-> setLinePen (QPen (Qt::black));
	Marker		-> attach (plotgrid);
	this		-> attach (plotgrid);
	IndexforMarker	= 0;

	lm_picker	= new QwtPlotPicker (plotgrid -> canvas ());
	lm_picker	-> setSelectionFlags (QwtPlotPicker::PointSelection);
	lm_picker	-> setMousePattern (QwtPlotPicker::MouseSelect1,
	                                    Qt::LeftButton);

	connect (lm_picker,
	         SIGNAL (selected (const QwtDoublePoint &)),
	         this,
	         SLOT (leftMouseClick (const QwtDoublePoint &)));
	plotgrid	-> replot();
}

	WaterfallViewer::~WaterfallViewer () {
	disconnect (lm_picker,
	            SIGNAL (selected (const QwtDoublePoint &)),
	            this,
	            SLOT (leftMouseClick (const QwtDoublePoint &)));
	plotgrid	-> enableAxis (QwtPlot::yRight, false);
	plotgrid	-> enableAxis (QwtPlot::xBottom, false);
	plotgrid	-> enableAxis (QwtPlot::yLeft, false);
	this		-> detach ();
	delete[]	plotData;
	delete		WaterfallData;
	delete		Marker;
	delete		lm_picker;
}

void	WaterfallViewer::leftMouseClick (const QwtDoublePoint & p) {
QPointF point = p;
	leftClicked ((int)(point. x ()) - IndexforMarker);
}

void	WaterfallViewer::ViewWaterfall (double *X_axis,
	                                double *Y1_value,
	                                double amp,
	                                int32_t marker) {
int32_t	orig	= (int32_t)(X_axis [0]);
int32_t	width	= (int32_t)(X_axis [Displaysize - 1] - orig);

	IndexforMarker	= marker;
	if (OneofTwo) {
	   OneofTwo = 0;
	   return;
	}
	OneofTwo	= 1;
/*
 *	shift one row, faster with memmove than writing out
 *	the loops. Note that source and destination overlap
 *	and we therefore use memmove rather than memcpy
 */
	memmove (&plotData [0],
	         &plotData [Displaysize], 
	         (Rastersize - 1) * Displaysize * sizeof (double));
/*
 *	... and insert the new line
 */
	memcpy (&plotData [(Rastersize - 1) * Displaysize],
	        &Y1_value [0],
	        Displaysize * sizeof (double));

	if (WaterfallData != NULL)
	   delete WaterfallData;

	WaterfallData = new SpectrogramData (plotData,
	                                     orig,
	                                     width,
	                                     Rastersize,
	                                     Displaysize,
	                                     amp);
	this		-> detach	();
	this		-> setData	(*WaterfallData);
	this		-> setDisplayMode (QwtPlotSpectrogram::ImageMode,
	                                                               TRUE);
	plotgrid	-> setAxisScale (QwtPlot::xBottom,
	                                 orig,
	                                 orig + width);
	plotgrid	-> enableAxis (QwtPlot::xBottom);

	Marker		-> setXValue  (marker);
	this		-> attach     (plotgrid);
	plotgrid	-> replot();
}
/*
 *	The spectrumDisplay
 */
	SpectrumViewer::SpectrumViewer (QwtPlot *plot, uint16_t displaysize) {

	plotgrid		= plot;
	this	-> Displaysize	= displaysize;
	plotgrid-> setCanvasBackground (Qt::white);
	grid	= new QwtPlotGrid;
	grid	-> setMajPen (QPen(Qt::gray, 0, Qt::DotLine));
	grid	-> enableXMin (true);
	grid	-> enableYMin (true);
	grid	-> setMinPen (QPen(Qt::gray, 0 , Qt::DotLine));
	grid	-> attach (plotgrid);

	SpectrumCurve	= new QwtPlotCurve ("");
   	SpectrumCurve	-> setPen (QPen(Qt::black));
//	SpectrumCurve	-> setBaseline	(0.0);
//	SpectrumCurve	-> setStyle	(QwtPlotCurve::Sticks);
	SpectrumCurve	-> attach (plotgrid);
	
	Marker		= new QwtPlotMarker ();
	Marker		-> setLineStyle (QwtPlotMarker::VLine);
	Marker		-> setLinePen (QPen (Qt::red));
	Marker		-> attach (plotgrid);
	plotgrid	-> enableAxis (QwtPlot::yLeft);

	lm_picker	= new QwtPlotPicker (plot -> canvas ());
	lm_picker	-> setSelectionFlags (QwtPlotPicker::PointSelection);
	lm_picker	-> setMousePattern (QwtPlotPicker::MouseSelect1,
	                                    Qt::LeftButton);

	rm_picker	= new QwtPlotPicker (plot -> canvas ());
	rm_picker	-> setSelectionFlags (QwtPlotPicker::PointSelection);
	rm_picker	-> setMousePattern (QwtPlotPicker::MouseSelect1,
	                                    Qt::RightButton);

	connect (lm_picker,
	         SIGNAL (selected (const QwtDoublePoint &)),
	         this,
	         SLOT (leftMouseClick (const QwtDoublePoint &)));
	connect (rm_picker,
	         SIGNAL (selected (const QwtDoublePoint &)),
	         this,
	         SLOT (rightMouseClick (const QwtDoublePoint &)));
	IndexforMarker	= 0;
}

	SpectrumViewer::~SpectrumViewer () {
	disconnect (lm_picker,
	            SIGNAL (selected (const QwtDoublePoint &)),
	            this,
	            SLOT (leftMouseClick (const QwtDoublePoint &)));
	disconnect (rm_picker,
	            SIGNAL (selected (const QwtDoublePoint &)),
	            this,
	            SLOT (rightMouseClick (const QwtDoublePoint &)));
	plotgrid	-> enableAxis (QwtPlot::yLeft, false);
	Marker		-> detach ();
	SpectrumCurve	-> detach ();
	grid		-> detach ();
	delete		Marker;
	delete		SpectrumCurve;
	delete		grid;
	delete		lm_picker;
}

void	SpectrumViewer::leftMouseClick (const QwtDoublePoint & p) {
QPointF point = p;
	leftClicked ((int)(point. x()) - IndexforMarker);
}

void	SpectrumViewer::rightMouseClick (const QwtDoublePoint &p) {
QPointF point = p;
	rightClicked ((int)(point. x()) - IndexforMarker);
}

void	SpectrumViewer::ViewSpectrum (double *X_axis,
		                      double *Y1_value,
	                              double amp,
	                              int32_t marker) {
uint16_t	i;

	amp		= amp / 100 * (-get_db (0));
	IndexforMarker	= marker;
	plotgrid	-> setAxisScale (QwtPlot::xBottom,
				         (double)X_axis [0],
				         X_axis [Displaysize - 1]);
	plotgrid	-> enableAxis (QwtPlot::xBottom);
	plotgrid	-> setAxisScale (QwtPlot::yLeft,
				         get_db (0), get_db (0) + amp);

	for (i = 0; i < Displaysize; i ++) 
	   Y1_value [i] = get_db (Y1_value [i]); 

	SpectrumCurve	-> setData (X_axis, Y1_value, Displaysize);
	Marker		-> setXValue (marker);
	plotgrid	-> replot(); 
}

