#
/*
 *
 *    Copyright (C)  2010, 2011, 2012
 *    Jan van Katwijk (J.vanKatwijk@gmail.com)
 *    JFF Consultancy
 *
 *    This file is part of the JSDR.
 *    Many of the ideas as implemented in JSDR are derived from
 *    other work, made available through the GNU general Public License. 
 *    All copyrights of the original authors are recognized.
 *
 *    JSDR is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    JSDR is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with JSDR; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef _GUI
#define _GUI

#include	"jff-include.h"
#include	<QDialog>
#include	"ui_sdrgui.h"
#include	<qwt.h>
#include	<qwt_slider.h>
#include	<qwt_plot.h>
#include	<qwt_plot_curve.h>
#include	<qwt_plot_marker.h>
#include	<qwt_plot_grid.h>
#include	<qwt_dial.h>
#include	<qwt_dial_needle.h>
#include	<qwt_plot_spectrogram.h>
#include	<QTimer>
#include	"scope.h"
#include	"virtual_input.h"
#include	<QKeyEvent>
#include	<QWheelEvent>
#include	"fft.h"
//
//	Starting with version 4.2 we include keyboard
//	alternatives for mouse clicks.
//	The keyboard filter is the generic handler for this

#include	<QObject>

class	QSettings;
class	Scope;
/*
 *	GThe main gui object. It inherits from
 *	QDialog and the generated form
 */
class RadioInterface: public QDialog,
		      private Ui_elektorSDR {
Q_OBJECT
public:
		RadioInterface		(QSettings	*,
	                                 QWidget *parent = NULL);
		~RadioInterface		();

private:

	Scope		*HFScope;
	int16_t		HFViewmode;
	DSPFLOAT	*Window;
	QSettings	*ISettings;
	bool		doInit;
	int32_t		CardRate;
	int32_t		segmentSize;
	int32_t		displaySize;
	int16_t		rasterSize;
	int32_t		spectrumSize;
	int16_t		segmentCount;
	int16_t		spectrumFactor;
	int32_t		scanDelay;
	bool		newSegment;
	int8_t		currentSegment;
	double		*X_axis;
	double		*displayVector;
	double		**displayBuffer;
	common_fft	*spectrum_fft;
	DSPCOMPLEX	*spectrumBuffer;
	virtualInput	*MyRig;
	float		currAttSliderValue;
	DSPFLOAT	attValueL;
	DSPFLOAT	attValueR;
	int64_t		currentFrequency;

	void		ClearPanel		();
	void		AddtoPanel		(int16_t);
	int		getPanel		();
	void		CorrectPanel		();

	void		Set_Tuner		(int32_t);
	void		Display			(int32_t, int16_t);
	QTimer		*lcd_timer;
	QTimer		*scanTimer;
	QTimer		*display_timer;
	QTimer		*runTimer;
	int32_t		frequency;
	int16_t		scopeAmplifier;
	int		Panel;
	int16_t		CurrentRig;
	bool		Running;
/*
 *	The UI acts as a kind of server, the public slots are the entries
 *	for the radio elements, signals (below) are the triggers
 *	for the radio elements.
 */
	void		IncrementFrequency	(int32_t);
	void		MaptoSegment		(uint8_t *, int32_t,
	                                         double *, int32_t);
/*
 */
private slots:
	void	stop_lcdTimer		(void);
	void	setStart		(void);
	void	lcd_timeout		(void);
	void	updateTimeDisplay	(void);

	void	AdjustFrequency		(int);
	void	addOne			(void);
	void	addTwo			(void);
	void	addThree		(void);
	void	addFour			(void);
	void	addFive			(void);
	void	addSix			(void);
	void	addSeven		(void);
	void	addEight		(void);
	void	addNine			(void);
	void	addZero			(void);
	void	addClear		(void);
	void	AcceptFreqinKhz		(void);
	void	addCorr			(void);

	void	setAttenuation		(int);
	void	setIQBalance		(int);

	void	setAmplification	(int);
	void	abortSystem		(int);
	void	TerminateProcess	(void);
	void	updateScanTimer		(void);
	void	setViewmode		(void);
public slots:
	void	JansRadioThroughSignal	(void);
};

#endif

