#
/*
 *
 *    Copyright (C) 2008, 2009, 2010
 *    Jan van Katwijk (J.vanKatwijk@gmail.com)
 *    JFF Consultancy
 *
 *    This file is part of the JFF SDR (JSDR).
 *    Many of the ideas as implemented in JSDR are derived from
 *    other work, made available through the GNU general Public License. 
 *    All copyrights of the original authors are recognized.
 *
 *    JSDR is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    JSDR is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with JSDR; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include	"decimator.h"
#include	<stdio.h>
/*
 *	A general down-decimator. Two parameters are
 *	a. samplerate in, 
 *	b. samplerate out,
 *
 *	if (in < out) it does not do anything
 *	if (out * N == in) it takes the simple road
 *	otherwise it does some counting and interpolating
 */

	downDecimator::downDecimator (int32_t inSamplerate,
	                              int32_t outSamplerate) {
	if  (inSamplerate < outSamplerate)
	   theWorker = new EDecimator (inSamplerate, outSamplerate);
	else
	if ((inSamplerate % outSamplerate) == 0)
	   theWorker = new IDecimator (inSamplerate, outSamplerate);
	else
	   theWorker = new FDecimator (inSamplerate, outSamplerate);
}

	downDecimator::~downDecimator (void) {
	delete theWorker;
}

bool	downDecimator::doDecimate (DSPCOMPLEX in, DSPCOMPLEX *out) {
	   return theWorker -> doDecimate (in, out);
}

bool	downDecimator::doDecimate (DSPFLOAT in, DSPFLOAT *out) {
	   return theWorker -> doDecimate (in, out);
}

	FDecimator::FDecimator (int32_t inSamplerate, int32_t outSamplerate) {
/*
 *	inperiod and outperiod are in picoseconds
 */
	this -> inSamplerate	= inSamplerate;
	this -> outSamplerate	= outSamplerate;
	inperiod		= (int64_t)(100000000000LL) / inSamplerate; 
	outperiod		= (int64_t)(100000000000LL) / outSamplerate;
	oldinsampleTime		= 0;
	currentinsampleTime	= 0;
	oldinsampleValue	= 0;
	currentinsampleValue	= 0;
	LPFilter		= new LowPassIIR (4,
                                                  (int)(0.5 * outSamplerate),
	                                          inSamplerate,
	                                          S_BUTTERWORTH);
	cycleCount	= 0;
}

	FDecimator::~FDecimator (void) {
	delete LPFilter;
}
/*
 *	This is the one
 */
bool	FDecimator::doDecimate (DSPCOMPLEX Sample, DSPCOMPLEX *newoutSample) {
double	ratio;
DSPCOMPLEX	newValue;

	oldinsampleTime		= currentinsampleTime;
	oldinsampleValue	= currentinsampleValue;
	currentinsampleTime	= oldinsampleTime + inperiod;
	currentinsampleValue	= Sample;

	if (currentinsampleTime < outperiod) 
	   return false;
	   
/*
 *	now we know that the output sample should have appeared on a time
 *	somewhere between oldinsampletime and currentinsampletime
 */
	ratio	= ((double)outperiod - oldinsampleTime) / (double)inperiod;

	newValue = oldinsampleValue +
	           cmul (currentinsampleValue - oldinsampleValue, ratio);
	*newoutSample = newValue;
//	*newoutSample = LPFilter -> Pass (newValue);

	currentinsampleTime	%= outperiod;
	return true;
}

bool	FDecimator::doDecimate (DSPFLOAT Sample, DSPFLOAT *newoutSample) {
DSPCOMPLEX temp;

	if (doDecimate (DSPCOMPLEX (Sample, 0), &temp)) {
	   *newoutSample = real (temp);
	   return true;
	}
	return false;
}

	IDecimator::IDecimator (int inSamplerate, int outSamplerate) {
	intRatio	= inSamplerate / outSamplerate;
	intCount	= intRatio;
}

	IDecimator::~IDecimator (void) {
}

bool	IDecimator::doDecimate (DSPCOMPLEX Sample, DSPCOMPLEX *outSample) {
	if (--intCount == 0) {
	   *outSample = Sample;
	   intCount = intRatio;
	   return true;
	}
	return false;
}

bool	IDecimator::doDecimate (DSPFLOAT Sample, DSPFLOAT *outSample) {
	if (-- intCount == 0) {
	   *outSample = Sample;
	   intCount = intRatio;
	   return true;
	}
	return false;
}


	EDecimator::EDecimator (int inSamplerate, int outSamplerate) {
	(void)inSamplerate;
	(void)outSamplerate;
}

	EDecimator::~EDecimator () {
}

bool	EDecimator::doDecimate (DSPCOMPLEX sample, DSPCOMPLEX *outSample) {
	*outSample = sample;
	return true;
}

bool	EDecimator::doDecimate (DSPFLOAT sample, DSPFLOAT *outSample) {
	*outSample = sample;
	return true;
}

	virtualDecimator::virtualDecimator () {
}

	virtualDecimator::~virtualDecimator () {
}

bool	virtualDecimator::doDecimate (DSPCOMPLEX Sample,
	                              DSPCOMPLEX *outSample) {
	*outSample = Sample;
	return true;
}

bool	virtualDecimator::doDecimate (DSPFLOAT Sample,
	                              DSPFLOAT *outSample) {
	*outSample = Sample;
	return true;
}

